import os
from unittest import TestCase

import torch

from pointcyto.io.loaders.fcs import read_fcs
from pointcyto.testutils.helpers import find_dirname_above

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above(THIS_DIR)

flowcyto_testdata_dir = os.path.join(TESTS_DIR, "testdata", "flowcytometry")
ultrasmall_fcs = os.path.join(flowcyto_testdata_dir, "ultra_small_fcs.fcs")

# The following code was used in R (project: 2019-09-14_facs_melanoma) to generate the exemplary data ultrasmall fcs
# library(flowCore)
# a <- read.FCS("Data/OperationalData/Tcell/P0H.fcs")
# exprs(a)
# exprs(a[1:10, ])
# write.FCS(a[1:3, 1:4], filename = "ultra_small_fcs.fcs")

target_tensor = torch.Tensor(
    [
        [15.83853, 20.82897, 15.71305, 8.452956],
        [13.82324, 20.64287, 13.55996, 6.616665],
        [16.00134, 20.70882, 15.53644, 8.520460],
    ]
)
target_tensor = target_tensor.double()

target_markernames = [
    "FS_INT_LIN_FS_INT",
    "FS_TOF_LIN_FS_TOF",
    "SS_INT_LIN_SS_INT",
    "CD45RA_FITC_CD45RA",
]


class Test(TestCase):
    def test_fcs_loader(self):
        if not os.path.exists(ultrasmall_fcs):
            self.fail(msg=ultrasmall_fcs + " does not exist")

        loaded_csv = read_fcs(ultrasmall_fcs)

        assert torch.allclose(target_tensor, loaded_csv[0])
        self.assertEqual(target_markernames, loaded_csv[1])



import copy
import pickle
import re
from typing import Any, Dict, List, Literal, Union

import pytorch_lightning as pl
import torch_geometric

from pointcyto.data.InMemoryPointCloud import InMemoryPointCloud
from pointcyto.data.metadata_base import MetaDataBase


class PCDataModule(pl.LightningDataModule):
    def __init__(
        self,
        meta_train: Dict[str, Union[MetaDataBase, str, Dict[str, Any]]] = None,
        meta_val: Dict[str, Union[MetaDataBase, str, Dict[str, Any]]] = None,
        meta_test: Dict[str, Union[MetaDataBase, str, Dict[str, Any]]] = None,
        meta_predict: Dict[str, Union[MetaDataBase, str, Dict[str, Any]]] = None,
        pointcloud_args_datapart: Dict[str, Dict[str, Any]] = None,
        pointcloud_generator=InMemoryPointCloud,
        batch_size: int = 43,
        num_workers: int = 0,
        must_prepare: bool = True,
        clear_processed: bool = False,
    ):
        """Pointcloud DataModule

        https://pytorch-lightning.readthedocs.io/en/latest/data/datamodule.html#lightningdatamodule-api

        Args:
            meta_XXX (Dict[str, Union[MetaDataBase str]], optional):
                `meta_["train", "val", "test", "predict"]`
                For tvtp-part, give a dictionary for the dataparts.
                If an element contains a _dictionary_, this is directly fed into `InMemoryPointcloud(**kwargs)` AND
                takes precedence over `pointcloud_args_datapart`!

                E.g.
                ```
                    meta_train = {
                        "dataA": "path_to_metadata_object",
                        "dataB": MetaDataBaseobject_B,
                        "dataC": {
                            "metadata": <path_OR_MetaDataObject>,
                            "pre_pre_transform_param_onlist": ...,
                            <further InMemoryPointCloud arguments>: <their values>
                        }
                    }

                Each element is finally brought into the format
                    "dataC": {
                        "metadata": <path_OR_MetaDataObject>,
                        "pre_pre_transform_param_onlist": ...,
                        <further InMemoryPointCloud arguments>: <their values>
                    }
                using `pointcloud_args_datapart`. If `pointcloud_args_datapart` is None, the
                result will be:
                    "dataC": {
                        "metadata": <path_OR_MetaDataObject>,
                    }
                and therefore an InMemoryPointCloud WITHOUT any parameters will be generated
                ```

                Defaults to None.
            pointcloud_args_datapart Dict[str, Dict[str, Any]]:
                If given, the datapart args
            pointcloud_generator (_type_, optional): _description_. Defaults to InMemoryPointCloud.
            batch_size (int, optional):
                Given to torch_geometric.loader.DataLoader(). Defaults to 43.
            num_workers (int, optional):
                Given to torch_geometric.loader.DataLoader(). Defaults to 0.
            must_prepare (bool, optional):
                When calling calling `PCDataModule.setup()`, is it mandatory to have called
                `PCDataModule.prepare_data()` first?

                Defaults to True.
            clear_processed (bool, optional):
                If True force all dataparts to have clear_processed = True therefore all PointClouds
                get reprocessed.
        """

        super().__init__()
        self.pointcloud_generator = pointcloud_generator
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.must_prepare = must_prepare
        self.meta_conf_xxx = {
            "train": copy.deepcopy(meta_train),
            "val": copy.deepcopy(meta_val),
            "test": copy.deepcopy(meta_test),
            "predict": copy.deepcopy(meta_predict),
        }
        dataparts = None
        for meta_key, meta_dataparts in self.meta_conf_xxx.items():
            if meta_dataparts is None:
                continue
            if not isinstance(meta_dataparts, dict):
                raise ValueError(
                    f"meta_{meta_key} should be a dictionary like"
                    + "{'dataA': metadata}"
                )
            # Check that all existing dataparts have the same datapart-names
            if dataparts is None:
                dataparts = set(meta_dataparts.keys())
            else:
                if dataparts != set(meta_dataparts.keys()):
                    raise ValueError(
                        f"Given {meta_key} contains different part-names than the existing ones."
                        + f"\nExisting: {*dataparts,}"
                        + f"\nNew ({meta_key}): {*list(meta_dataparts.keys()),}"
                    )

            # Generate the dictionary containing the metadata paths and arguments for InMemoryPointCloud
            # "dp" := "datapart"
            for dp_name, dp_value in meta_dataparts.items():
                if isinstance(dp_value, dict):
                    if "metadata" in dp_value.keys():
                        # Then we have a defined PointCloud already, do nothing
                        pass
                    else:
                        # Then we have a defined PointCloud already, but the metadata
                        # (where to load the data from) is missing
                        raise ValueError(
                            f"Given {meta_key}:{dp_name} is a dict but does not contain 'metadata' as key (It HAS to)."
                        )
                else:
                    if isinstance(dp_value, str):
                        # Then the current dp_value is
                        #   "dataA": "path_to_metadata_object",
                        with open(dp_value, "rb") as f:
                            metadata = pickle.load(f)
                    elif isinstance(dp_value, MetaDataBase):
                        #   "dataA": MetaDataBaseobject_B,
                        metadata = dp_value
                    else:
                        raise ValueError(
                            f"Given {meta_key}:{dp_name} is neither a dict, nor a string, not a MetaDataBase object."
                        )
                    new_dp_value = {"metadata": metadata}
                    if pointcloud_args_datapart is not None:
                        # pointcloud_args_datapart: Dict[str, Dict[str, Any]]
                        # {
                        #     "dataA": {
                        #         "transform": FixedPoints(100, replace=True),
                        #     },
                        # }
                        if set(pointcloud_args_datapart.keys()) != dataparts:
                            raise ValueError(
                                "pointcloud_args_datapart had different keys than dataparts:"
                                + f"\n{set(pointcloud_args_datapart.keys())}\n{dataparts}\n"
                            )
                        new_dp_value.update(pointcloud_args_datapart[dp_name])
                    self.meta_conf_xxx[meta_key][dp_name] = new_dp_value
        # After this initialization,
        # `self.meta_xxx` is a dictionary where for every element (train, val, test, predict)
        # it looks as following:
        #   {
        #     "dataA": {
        #         "metadata": MetaDataObject,
        #         <further InMemoryPointCloud arguments>: <values>
        #     },
        #     "dataB": {
        #         "metadata": MetaDataObject,
        #         <further InMemoryPointCloud arguments>: <values>
        #     }
        #   }

        if clear_processed:
            # Then force all dataparts to have clear_processed = True
            # therefore all PointClouds get reprocessed
            for meta_key, meta_dataparts in self.meta_conf_xxx.items():
                if meta_dataparts is not None:
                    # "dp" := "datapart"
                    for dp_name, dp_value in meta_dataparts.items():
                        self.meta_conf_xxx[meta_key][dp_name]["clear_processed"] = True

        # pointcloud_xxx (xxx = train, val, test, predict)
        # Is not necessarily InMemoryPointCloud but could be usual PointCloud
        self.pc_xxx: Dict[
            Literal["train", "val", "test", "predict"], InMemoryPointCloud
        ] = {}

    def prepare_data(self, clear_processed: bool = False):
        """Prepare the data

        Downloading and preprocessing. Usually set up all given data sources with InMemoryPointCloud.

        From pytorch-lightning documentation:
            prepare_data is called from the main process. It is not recommended to assign state here
            (e.g. self.x = y) since it is called on a single process and if you assign states here
            then they won't be available for other processes.
        """
        for tvt_name, tvt in self.meta_conf_xxx.items():
            if tvt is not None:
                # "dp" := "datapart"
                for dp_name, dp_value in tvt.items():
                    if clear_processed:
                        dp_value["clear_processed"] = True
                    self.pointcloud_generator(**dp_value)
                    # If `clear_processed` was True in the beginning, after the data was initially
                    # loaded (=prepared), this should not be loaded over and over again, therefore
                    # set `clear_processed` to False
                    self.meta_conf_xxx[tvt_name][dp_name].update(
                        {"clear_processed": False}
                    )
        self.must_prepare = False

    def _stage_interpreter(
        self, stage: Literal["fit", "validate", "test", "predict"] = None
    ) -> List[Literal["train", "val", "test", "predict"]]:
        if stage in ["train", "val", "test", "predict"]:
            return [stage]

        if stage is None:
            stage_list = list(self.meta_conf_xxx.keys())
        else:
            if stage == "fit":
                stage_list = ["train", "val"]
            elif stage == "validate":
                stage_list = ["val"]
            elif stage in ["test", "predict"]:
                stage_list = [stage]
            else:
                raise ValueError(f"DataModule: I do not know how to .setup() {stage}")
        return stage_list

    def setup(self, stage: Literal["fit", "validate", "test", "predict"] = None):
        if self.must_prepare:
            raise ValueError(
                "self.must_prepare is True, call PCDataModule.prepare_data() before .setup()"
            )
        stage_list = self._stage_interpreter(stage)

        for stage in stage_list:
            if self.meta_conf_xxx[stage] is not None:
                self.pc_xxx[stage] = {}
                for dp_name, dp_value in self.meta_conf_xxx[stage].items():
                    self.pc_xxx[stage][dp_name] = self.pointcloud_generator(**dp_value)

    def _part_dataloader(self, tvtp: Literal["train", "val", "test", "predict"]):
        try:
            return pl.utilities.combined_loader.CombinedLoader(
                {
                    datapart: torch_geometric.loader.DataLoader(
                        pc_xxx_datapart,
                        batch_size=self.batch_size,
                        shuffle=False,
                        num_workers=self.num_workers,
                    )
                    for datapart, pc_xxx_datapart in self.pc_xxx[tvtp].items()
                }
            )
        except KeyError:
            raise KeyError(f"Did you call PCDataModule.setup({tvtp})?")

    def train_dataloader(self):
        return self._part_dataloader("train")

    def val_dataloader(self):
        return self._part_dataloader("val")

    def test_dataloader(self):
        return self._part_dataloader("test")

    def predict_dataloader(self):
        return self._part_dataloader("predict")

    def __repr__(self) -> str:
        rstr = ""
        for tvtp, tvtp_values in self.meta_conf_xxx.items():
            rstr += f"\n\n{tvtp}\n"
            if tvtp_values is not None:
                for dp_name, dp_value in tvtp_values.items():
                    rstr += f"\n  {dp_name}" + re.sub(
                        r"\n", r"\n     ", r"\n" + dp_value.__repr__() + "\n"
                    )

        return rstr



import re
from typing import List, Union

import numpy as np
import torch
from torch_geometric.data.dataset import Dataset

from pointcyto.transforms.utils import standard_repr


def global_mean(pointclouds: Dataset, dataset_attribute: str = "pos") -> torch.Tensor:
    """
    Calculate the mean over all points in all samples of pointclouds

    Args:
        pointclouds:
            A torch_geometric.data.dataset.
        dataset_attribute:
            The attribute of Dataset where the mean will be taken

    Returns:
        A torch.tensor holding the calculated means over all samples per feature

    """
    means = torch.stack(
        [getattr(sample, dataset_attribute).mean(dim=0) for sample in pointclouds]
    )
    overall_mean = means.mean(dim=0)
    return overall_mean


def global_var(
    pointclouds: Dataset,
    dataset_attribute: str = "pos",
    global_means: torch.Tensor = None,
) -> torch.Tensor:
    """
    Calculate the variance over all points in all samples of pointclouds

    Args:
        pointclouds:
            A torch_geometric.data.dataset.
        dataset_attribute:
            The attribute of Dataset where the variance will be calculated
        global_means:
            If the global mean was calculated before, the result of it can be given to global_var

    Returns:
        A torch.tensor holding the calculated variance over all samples per feature

    """
    if global_means is not None:
        mean_overall = global_means
    else:
        mean_overall = global_mean(pointclouds, dataset_attribute)
    x_minus_global_mean = torch.cat(
        [getattr(sample, dataset_attribute) - mean_overall for sample in pointclouds]
    )
    x_minus_g_squared = x_minus_global_mean**2
    summed_squares = x_minus_g_squared.sum(dim=0)
    return summed_squares / x_minus_global_mean.shape[0]


def global_sd(
    pointclouds: Dataset,
    dataset_attribute: str = "pos",
    global_means: torch.Tensor = None,
) -> torch.Tensor:
    """
    Calculate the standard deviation over all points in all samples of pointclouds

    Args:
        pointclouds:
            A torch_geometric.data.dataset.
        dataset_attribute:
            The attribute of Dataset where the standard deviation will be taken
        global_means:
            If the global mean was calculated before, the result of it can be given to global_var

    Returns:
        A torch.tensor holding the calculated standard deviation over all samples per feature

    """
    return global_var(pointclouds, dataset_attribute, global_means).sqrt()


def functional_transform_adapter(
    data: Dataset, transform, dataset_attribute: str = "pos", *args, **kwargs
) -> Dataset:
    """
    To be able to use existing transforms on the :class:`torch_geometric.data.dataset.Dataset` I need this adapter.
    You have to defined the data, the transform and on which attribute of the given dataset the transform should be
    applied. *args and **kwargs go into the transform.

    Args:
        data:
            The :class:`torch_geometric.data.dataset.Dataset` where you want to apply the transform on.
        transform:
            The transform you want to apply.
        dataset_attribute:
            The attribute of :class:`torch_geometric.data.dataset.Dataset` where you want to apply the transform on.
        *args and **kwargs:
            Further arguments for ``transform``

    Returns:
        The transformed dataset_attribute of :class:`torch_geometric.data.dataset.Dataset` (not a dataset anymore!).

    """
    return transform(getattr(data, dataset_attribute), *args, **kwargs)


class NormalizePointCloud(object):
    def __init__(
        self, mean, std, inplace: bool = False, dataset_attribute: str = "pos"
    ):
        """

        Args:
            mean:
                1D-Tensor with the means for each column.
                The i-th value is subtracted from all values of the i-th column of the matrix given by
                dataset_attribute.
            std:
                 1D-Tensor with the standard deviations for each column.
                The i-th value is subtracted from all values of the i-th column of the matrix given by
                dataset_attribute.
            inplace:
                Calculation in place or clone the tensor first?
            dataset_attribute:
                Data from pytorch-geometric has multiple data matrices where you can work on. Default for us: "pos"
        """
        self.mean = mean
        self.std = std
        self.inplace = inplace
        self.attr = dataset_attribute

    def __call__(self, data: Dataset) -> Dataset:
        """
        Normalizes the ``dataset_attribute`` from the given dataset.
        ((x - mean(x)) / sd(x)) per feature using ``self.mean`` and ``self.std`` over all samples.
        """
        setattr(
            data,
            self.attr,
            functional_transform_adapter(
                data, normalize, self.attr, self.mean, self.std, self.inplace
            ),
        )
        return data

    def __repr__(self):
        return standard_repr(self)


def normalize(tensor, mean, std, inplace=False) -> torch.Tensor:
    """Normalize a tensor with mean and standard deviation.

    .. note::
        This transform acts out of place by default, i.e., it does not mutates the input tensor.

    See :class:`~torchvision.transforms.Normalize` for more details.

    Args:
        tensor (Tensor):
            Tensor image of size (C, H, W) to be normalized.
        mean (sequence):
            Sequence of means for each channel.
        std (sequence):
            Sequence of standard deviations for each channel.
        inplace(bool,optional):
            Bool to make this operation inplace.

    Returns:
        Tensor: Normalized Tensor image.
    """

    if not inplace:
        tensor = tensor.clone()

    dtype = tensor.dtype
    mean = torch.as_tensor(mean, dtype=dtype, device=tensor.device)
    std = torch.as_tensor(std, dtype=dtype, device=tensor.device)
    tensor.sub_(mean).div_(std)
    return tensor


class SelectFeaturesPointCloud(object):
    def __init__(
        self,
        features: Union[List[int], slice, int],
        original_featurenames: List[str] = None,
        inplace: bool = False,
        dataset_attribute: str = "pos",
    ):
        """
        Restrict the data.dataset_attribute to the specified features.

        Args:
            features:
                The features which should be present *after* the selection. Define via integers, List[int] or a slice.
            original_featurenames:
                All features **before** restriction. If given, ``self.featurenames`` holds the restricted featurenames
                (which are based on index) already after initialization of the class, not only after calling it.
            inplace:
                Inplace or not.
            dataset_attribute:
                Which attribute should this function be applied to.

        """
        self.attr = dataset_attribute
        self.inplace = inplace
        self.orig_featurenames = original_featurenames
        self.feature_indices = features
        if self.orig_featurenames is not None:
            self.featurenames = [
                self.orig_featurenames[i] for i in self.feature_indices
            ]
        else:
            self.featurenames = None

    def __call__(self, data: Dataset) -> Dataset:
        setattr(
            data,
            self.attr,
            functional_transform_adapter(
                data, select_features, self.attr, self.feature_indices, self.inplace
            ),
        )
        return data


def select_features(tensor, col_indices, inplace=False):
    """
    Select certain columns of a 2d tensor

    .. note::
        This transform acts out of place by default, i.e., it does not mutates the input tensor.

    Args:
        tensor (Tensor): Tensor image of size (H, W) where col_indices from W are selected
        col_indices (index): Indices of the columns to retain
        inplace(bool,optional): Bool to make this operation inplace.

    Returns:
        Tensor: Normalized Tensor image.
    """
    if len(tensor.shape) != 2:
        raise ValueError("tensor.shape should have len 2")
    if not inplace:
        tensor = tensor.clone()
    return tensor[:, col_indices]


# Heavily based on
# https://pytorch-geometric.readthedocs.io/en/latest/_modules/torch_geometric/transforms/fixed_points.html#FixedPoints
class WeightedFixedPoints(object):
    r"""
    Samples a fixed number of :obj:`num` points and features from a point
    cloud where each point receives a weight to be sampled.

    Args:
        num (int):
            The number of points to sample.
        n_reference:
            The number of points which are sampled randomly.
            Those num_selection points are used as base.
            Outlierness-score is calculated relative to this random selection.
        data_selection:
            The torch_geometric.data.data.Data attribute which should be used for outlierness-calculation

        replace (bool, optional):
            Sample with (True) or without (False) replacement
    """

    def __init__(
        self,
        num: int,
        n_reference: int,
        replace: bool = True,
        data_selection: str = "pos",
        percentage_outlierness_points: float = 0.5,
    ):
        self.num = num
        self.num_selection = n_reference
        self.data_selection = data_selection
        self.percentage_outlierness_points = percentage_outlierness_points
        self.replace = replace

    def __call__(self, data):
        num_nodes = data.num_nodes

        # Get the random points
        if self.num_selection > num_nodes:
            current_num_selection = num_nodes
        else:
            current_num_selection = self.num_selection
        random_point_choice = np.random.choice(
            a=num_nodes, size=current_num_selection, replace=False
        )
        random_points = getattr(data, self.data_selection)[
            random_point_choice, :
        ].cuda()

        # Calculate outlierness-score
        distance_randompoints_to_allpoints: torch.Tensor = torch.cdist(
            random_points, getattr(data, self.data_selection).cuda()
        )
        outlierness_score = distance_randompoints_to_allpoints.min(dim=0)[
            0
        ]  # ['values']
        outlierness_score = outlierness_score.cpu().numpy()
        outlierness_probabilites = outlierness_score / outlierness_score.sum()

        # Sample indices randomly (with/without replacement) with weights from all points
        # Higher outlierness-score means higher probability to be chosen.
        n_outlierness_points = int(self.num * self.percentage_outlierness_points)
        n_random_points = self.num - n_outlierness_points

        choice_outlierness = np.random.choice(
            a=num_nodes,
            size=n_outlierness_points,
            p=outlierness_probabilites,
            replace=self.replace,
        )
        choice_random = np.random.choice(
            a=num_nodes, size=n_random_points, replace=self.replace
        )
        choice = np.concatenate([choice_outlierness, choice_random])

        # Actually sample the data.
        for key, item in data:
            if bool(re.search("edge", key)):
                continue
            if torch.is_tensor(item) and item.size(0) == num_nodes and num_nodes != 1:
                data[key] = item[choice]

        return data

    def __repr__(self):
        return standard_repr(self)


class CalculateOutlierness(object):
    def __init__(
        self,
        n_reference: int,
        from_attribute: str = "pos",
        new_attribute: str = "outlierness",
        use_cuda: bool = True,
        calculate_in_batches: int = None,
    ):
        self.from_attr = from_attribute
        self.new_attr = new_attribute
        self.n_reference = n_reference
        self.use_cuda = use_cuda
        self.calculate_in_batches = calculate_in_batches

    def __call__(self, data: Dataset):
        """
        Args:

        Returns:
            Tensor: Normalized Tensor image.
        """
        setattr(
            data,
            self.new_attr,
            functional_transform_adapter(
                data,
                calculate_outlierness,
                self.from_attr,
                self.n_reference,
                self.use_cuda,
                self.calculate_in_batches,
            ),
        )
        return data

    def __repr__(self):
        return standard_repr(self)


def calculate_outlierness(
    tensor: torch.Tensor,
    n_reference: int,
    use_cuda: bool = True,
    cdist_max_points: int = None,
) -> torch.Tensor:
    # Get the random points
    if n_reference > tensor.shape[0]:
        current_num_selection = tensor.shape[0]
    else:
        current_num_selection = n_reference
    random_point_choice = np.random.choice(
        a=tensor.shape[0], size=current_num_selection, replace=False
    )
    random_points = tensor[random_point_choice, :]
    if use_cuda:
        random_points = random_points.cuda()

    # Calculate outlierness-score
    if cdist_max_points is not None and cdist_max_points > 1:
        num_batches = tensor.shape[0] // cdist_max_points
        outlierness_score = torch.tensor([], dtype=torch.float)
        if use_cuda:
            outlierness_score = outlierness_score.cuda()
        for batch_i in range(
            1, num_batches + 1 + (tensor.shape[0] % cdist_max_points != 0)
        ):
            end = batch_i * cdist_max_points
            if end > tensor.shape[0]:
                end = tensor.shape[0]
            batch_slice = slice((batch_i - 1) * cdist_max_points, end)
            part_tensor = tensor[batch_slice, :]
            if use_cuda:
                part_tensor = part_tensor.cuda()
            distance_randompoints_to_part_tensor: torch.Tensor = torch.cdist(
                random_points, part_tensor
            )
            part_outlierness_score = distance_randompoints_to_part_tensor.min(dim=0)[
                0
            ]  # ['values']
            outlierness_score = torch.cat([outlierness_score, part_outlierness_score])
    else:
        if use_cuda:
            tensor = tensor.cuda()
        distance_randompoints_to_allpoints: torch.Tensor = torch.cdist(
            random_points, tensor
        )
        outlierness_score = distance_randompoints_to_allpoints.min(dim=0)[
            0
        ]  # ['values']

    # If I had to set cdist_max_points, I most probable also have to convert the gpu-vector to cpu because of
    # GPU overflow.
    # This seems to take longer but better than not working at all.
    if cdist_max_points is not None:
        return outlierness_score.cpu()
    else:
        return outlierness_score


class FixedPointsOutlierness(object):
    r"""Samples a fixed number of :obj:`num` points and features from a point
    cloud where the Data has an 'outlierness' attribute which accounts for the outlierness of each point.

    Args:
        num (int):
            The number of points to sample.
        outlierness_attribute:
            The torch_geometric.data.data.Data attribute which should be used as outlierness-score
        replace (bool, optional):
            Sample with (True) or without (False) replacement
        percentage_outlierness_points:
            Percentage of points which are sampled according to outlierness-score probabilities.
        outlierness_exponent:
            You might want to weight distribute your outlierness score more.
            So just take all values to some exponent > 1 before calculating the probabilities.
            The other way around if your outlierness_exponent is < 1.
            Extreme cases:
                0   -> all points randomly sampled
                Inf -> Only the point with maximum outlierness score is sampled.
        allow_duplicates (bool, optional): In case :obj:`replace` is
            :obj`False` and :obj:`num` is greater than the number of points,
            this option determines whether to add duplicated nodes to the
            output points or not.
            In case :obj:`allow_duplicates` is :obj:`False`, the number of
            output points might be smaller than :obj:`num`.
            In case :obj:`allow_duplicates` is :obj:`True`, the number of
            duplicated points are kept to a minimum. (default: :obj:`False`)
    """

    def __init__(
        self,
        num: int,
        outlierness_attribute: str = "outlierness",
        replace: bool = True,
        percentage_outlierness_points: float = 0.5,
        outlierness_exponent: float = 1.5,
        allow_duplicates: bool = False,
    ):
        self.num = num
        self.outlierness_attribute = outlierness_attribute
        self.percentage_outlierness_points = percentage_outlierness_points
        self.outlierness_exponent = outlierness_exponent
        self.replace = replace
        self.allow_duplicates = allow_duplicates

    def __call__(self, data):
        num_nodes = data.num_nodes
        if self.num > num_nodes and not self.replace:
            if not self.allow_duplicates:
                raise ValueError(
                    "There are less points than defined to sample "
                    + "in num. Also allow_duplicates was set to false"
                )
            else:
                choice_shuffled = np.random.choice(
                    a=num_nodes, size=num_nodes, replace=False
                )
                choice_duplicates = np.random.choice(
                    a=num_nodes, size=self.num - num_nodes, replace=True
                )
                choice = np.concatenate([choice_shuffled, choice_duplicates])
        else:
            # Calculate outlierness-probabilities
            outlierness_score = data[self.outlierness_attribute].cpu().numpy()
            exponented = outlierness_score**self.outlierness_exponent
            outlierness_probabilites = exponented / exponented.sum()

            # Sample indices randomly (with/without replacement) with weights from all points
            # Higher outlierness-score means higher probability to be chosen.
            # print(
            #     self.num,
            #     self.percentage_outlierness_points,
            #     self.num * self.percentage_outlierness_points,
            #     int(self.num * self.percentage_outlierness_points),
            # )
            n_outlierness_points = int(self.num * self.percentage_outlierness_points)
            n_random_points = self.num - n_outlierness_points

            choice_outlierness = np.random.choice(
                a=num_nodes,
                size=n_outlierness_points,
                p=outlierness_probabilites,
                replace=self.replace,
            )
            choice_random = np.random.choice(
                a=num_nodes, size=n_random_points, replace=self.replace
            )
            choice = np.concatenate([choice_outlierness, choice_random])

        # Actually sample the data.
        for key, item in data:
            if bool(re.search("edge", key)):
                continue
            if torch.is_tensor(item) and item.size(0) == num_nodes and num_nodes != 1:
                data[key] = item[choice]

        return data

    def __repr__(self):
        return standard_repr(self)


class QminmaxPointCloud(object):
    def __init__(
        self,
        quantile_range=[0.01, 0.99],
        inplace: bool = False,
        dataset_attribute: str = "pos",
    ):
        """

        Args:
            quantile_range (sequence):
                Two quantiles reflecting the upper and lower quantile to be used as min and max replacements.
            inplace:
                Calculation in place or clone the tensor first?
            dataset_attribute:
                Data from pytorch-geometric has multiple data matrices where you can work on. Default for us: "pos"
        """
        self.quantile_range = quantile_range
        self.inplace = inplace
        self.attr = dataset_attribute

    def __call__(self, data: Dataset) -> Dataset:
        """
        Normalizes the ``dataset_attribute`` from the given dataset.
        ((x - upper_quantile(x)) / (upper_quantile(x) - lower_quantile(x))) per feature
        """
        setattr(
            data,
            self.attr,
            functional_transform_adapter(
                data, qminmax, self.attr, self.quantile_range, self.inplace
            ),
        )
        return data

    def __repr__(self):
        return standard_repr(self)


def qminmax(tensor: torch.Tensor, quantile_range, inplace=False) -> torch.Tensor:
    """
    minmax rescaling a tensor per column,
    (https://en.wikipedia.org/wiki/Feature_scaling#Rescaling_(min-max_normalization))
    but instead of using min and max use min := lower_quantile and max := upper_quantile from quantile_range

    .. note::
        This transform acts out of place by default, i.e., it does not mutates the input tensor.

    See :class:`~torchvision.transforms.Normalize` for more details.

    Args:
        tensor (Tensor):
            Tensor image of size (C, H, W) to be normalized.
        quantile_range (sequence):
            Two quantiles reflecting the upper and lower quantile to be used as min and max replacements.
        inplace(bool,optional):
            Bool to make this operation inplace.

    Returns:
        Tensor: Normalized Tensor image.
    """

    if not inplace:
        tensor = tensor.clone()

    dtype = tensor.dtype
    quantile_range = torch.as_tensor(quantile_range, dtype=dtype, device=tensor.device)
    upper_quantile = tensor.quantile(q=torch.max(quantile_range), dim=0, keepdim=True)
    lower_quantile = tensor.quantile(q=torch.min(quantile_range), dim=0, keepdim=True)
    return (tensor - lower_quantile) / (upper_quantile - lower_quantile)



"""
    Dummy conftest.py for nbnode.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    - https://docs.pytest.org/en/stable/fixture.html
    - https://docs.pytest.org/en/stable/writing_plugins.html
"""
# https://stackoverflow.com/questions/17801300/how-to-run-a-method-before-all-tests-in-all-classes


def pytest_configure(config):
    """
    Allows plugins and conftest files to perform initial configuration.
    This hook is called for every plugin and initial conftest
    file after command line options have been parsed.
    """


def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    import shutil

    shutil.rmtree("tests_output", ignore_errors=True)


def pytest_sessionfinish(session, exitstatus):
    """
    Called after whole test run finished, right before
    returning the exit status to the system.
    """


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """



import os
import unittest

import torch

from pointcyto.io.loaders.csv import read_csv
from pointcyto.testutils.helpers import find_dirname_above

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above(THIS_DIR)

flowcyto_testdata_dir = os.path.join(TESTS_DIR, "testdata", "flowcytometry")

target_tensor = torch.tensor([[1.0, 1.0, 1.0, 1.0], [13.0, 2.0, 4.0, 12.0]])
target_tensor = target_tensor.double()

target_markernames = ["Marker1", "Marker2", "Marker3", "Marker4"]


class Test(unittest.TestCase):
    def test_is_testdata_present(self):
        if not os.path.exists(flowcyto_testdata_dir):
            self.fail(msg=flowcyto_testdata_dir + " does not exist")

    def test_csv_loader(self):
        single_csv_file = os.path.join(
            flowcyto_testdata_dir, "Tcell_csv", "ultra_small_test_sample.csv"
        )
        if not os.path.exists(single_csv_file):
            self.fail(msg=single_csv_file + " does not exist")

        loaded_csv = read_csv(single_csv_file)

        assert torch.all(torch.eq(target_tensor, loaded_csv[0]))
        self.assertEqual(target_markernames, loaded_csv[1])



from typing import Any, Dict

from pointcyto.io.loaders.csv import read_csv
from pointcyto.io.loaders.fcs import read_fcs
from pointcyto.io.loaders.feather import read_feather
from pointcyto.io.loaders.jay import read_jay
from pointcyto.io.loaders.pt import read_pt
from pointcyto.io.utils import parse_matrix_class

POINTCLOUD_EXTENSIONS = [".fcs", ".csv", ".jay", ".pt", ".feather"]


def parse_by_ext(path: str, y: int = 0, *args, **kwargs) -> Dict[str, Any]:
    """
    Parse (=make a torch_geometric.data.Data object) file at <path> based on its extension.

    Args:
        path:
            Path to file
        y:
            What is the class response? Necessary for parse_matrix_class. Could also be left empty and set afterwards.
        *args:
        **kwargs:

    Returns:
        A Dict:
        ``point_data``:

            The torch_geometric.data.Data object

        ``features``:

            The feature names of the file
    """
    lowpath = path.lower()
    if lowpath.endswith(".jay"):
        point_set, col_names = read_jay(path, *args, **kwargs)
    elif lowpath.endswith(".pt"):
        point_set, col_names = read_pt(path, *args, **kwargs)
    elif lowpath.endswith(".csv"):
        point_set, col_names = read_csv(
            path, *args, **kwargs
        )  # further arguments go inside pandas.read_csv
    elif lowpath.endswith(".fcs"):
        point_set, col_names = read_fcs(path)
    elif lowpath.endswith(".feather"):
        point_set, col_names = read_feather(path, *args, **kwargs)
    else:
        raise NotImplementedError(
            "I can only handle the following extensions: "
            + ", ".join(POINTCLOUD_EXTENSIONS)
            + "\nYou gave the following file: "
            + lowpath
        )

    point_data = parse_matrix_class(point_set, y)
    return {"point_data": point_data, "features": col_names}



import os
from unittest import TestCase

import torch

from pointcyto.io.loaders.csv import read_csv
from pointcyto.io.loaders.fcs import read_fcs
from pointcyto.io.loaders.feather import read_feather
from pointcyto.io.meta_read_foldering import gen_csv_meta
from pointcyto.io.utils import parse_matrix_class
from pointcyto.testutils.helpers import find_dirname_above

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above(THIS_DIR)


pointcloud_toy_dataset = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_dataset"
)
pointcloud_toy_dataset_csvpheno = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_phenocsv"
)
tcell_dataset_singlefile = os.path.join(
    "testdata", "flowcytometry", "Tcell_csv", "ultra_small_test_sample.csv"
)


class Test(TestCase):
    def test_manual_readin_and_data_convertion(self):
        tmp = gen_csv_meta(
            os.path.realpath(os.path.join(TESTS_DIR, pointcloud_toy_dataset_csvpheno)),
            path_phenodata="samples_pheno.csv",
        )
        for single_sample in tmp:
            print(single_sample)
            single_fcs, colnames = read_fcs(
                os.path.join(tmp.orig_dir, single_sample["raw_filenames"])
            )
            myclass = parse_matrix_class(
                point_position_matrix=single_fcs, class_id=single_sample["y"]
            )
            print(myclass)
            print(myclass.pos)
            break

    def test_readcsv(self):
        read_tcell_file = read_csv(os.path.join(TESTS_DIR, tcell_dataset_singlefile))
        print(read_tcell_file)
        assert torch.equal(
            read_tcell_file[0],
            (
                torch.tensor(
                    [
                        [
                            1,
                            1,
                            1,
                            1,
                        ],
                        [13, 2, 4, 12],
                    ],
                    dtype=torch.int32,
                )
            ),
        )
        assert read_tcell_file[1] == ["Marker1", "Marker2", "Marker3", "Marker4"]

    def test_readcsv_datatable(self):
        import datatable

        def csv_reader(*args, **kwargs):
            read_data = datatable.fread(*args, **kwargs)
            return torch.tensor(read_data.to_numpy()), list(read_data.names)

        read_tcell_file = csv_reader(os.path.join(TESTS_DIR, tcell_dataset_singlefile))
        print(read_tcell_file)
        assert torch.equal(
            read_tcell_file[0],
            (
                torch.tensor(
                    [
                        [
                            1,
                            1,
                            1,
                            1,
                        ],
                        [13, 2, 4, 12],
                    ],
                    dtype=torch.int32,
                )
            ),
        )
        assert read_tcell_file[1] == ["Marker1", "Marker2", "Marker3", "Marker4"]

    def test_readcsv_pandas(self):
        import pandas as pd

        # fallback to pandas
        def csv_reader(*args, **kwargs):
            print("I am inside the fallback function")
            read_data = pd.read_csv(*args, **kwargs)
            return torch.tensor(read_data.values), [
                x.strip() for x in read_data.columns
            ]

        read_tcell_file = csv_reader(os.path.join(TESTS_DIR, tcell_dataset_singlefile))
        print(read_tcell_file)
        assert torch.equal(
            read_tcell_file[0],
            (
                torch.tensor(
                    [
                        [
                            1,
                            1,
                            1,
                            1,
                        ],
                        [13, 2, 4, 12],
                    ],
                    dtype=torch.int64,
                )
            ),
        )
        assert read_tcell_file[1] == ["Marker1", "Marker2", "Marker3", "Marker4"]

    def test_read_feather_pyarrow(self):
        # Run within R:
        #   tmp <- data.table::fread("tests/testdata/flowcytometry/Tcell_csv/ultra_small_test_sample.csv")
        #   dir.create("tests/testdata/flowcytometry/Tcell_feather")
        #   feather::write_feather(tmp, "tests/testdata/flowcytometry/Tcell_feather/ultra_small_test_sample.feather")
        read_tcell_file = read_feather(
            os.path.join(
                TESTS_DIR,
                "testdata",
                "flowcytometry",
                "Tcell_feather",
                "ultra_small_test_sample.feather",
            )
        )
        print(read_tcell_file)
        assert torch.equal(
            read_tcell_file[0],
            (
                torch.tensor(
                    [
                        [
                            1,
                            1,
                            1,
                            1,
                        ],
                        [13, 2, 4, 12],
                    ],
                    dtype=torch.int32,
                )
            ),
        )
        assert read_tcell_file[1] == ["Marker1", "Marker2", "Marker3", "Marker4"]



import os
import pathlib
import shutil
import warnings
from unittest import TestCase

import torch
import torch_geometric as tgeo

# from torch.utils.data.dataloader import DataLoader as t_DataLoader
from torch_geometric.loader import DataLoader as tg_DataLoader
from torch_geometric.transforms.fixed_points import FixedPoints

from pointcyto.data.InMemoryPointCloud import InMemoryPointCloud
from pointcyto.io.meta_read_foldering import gen_foldering_meta, gen_meta_filenames
from pointcyto.testutils.helpers import find_dirname_above

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above(THIS_DIR)

pointcloud_toy_dataset = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_dataset"
)
tcell_toy_dataset = os.path.join("testdata", "flowcytometry", "Tcell_foldering")
empty_sample_dataset = os.path.join("testdata", "flowcytometry", "empty_sample")


class TestInMemoryPointCloud(TestCase):
    def test_InMemoryPointCloud(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        b = InMemoryPointCloud(metadata=mymeta)
        assert b.metadata == a.metadata
        assert torch.equal(b.pos, a.pos)
        assert a != b

        shutil.rmtree(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "autoroot_raw"))
        # Now the dataset must be rebuilt
        c = InMemoryPointCloud(metadata=mymeta)
        assert c.metadata == a.metadata
        assert torch.equal(c.pos, a.pos)
        assert a != c
        repr(a)

    def test_InMemoryPointCloud_regression(self):
        from pointcyto.io.meta_read_foldering import gen_csv_meta

        pointcloud_toy_dataset_csvpheno = os.path.join(
            "testdata", "flowcytometry", "PointClouds_toy_phenocsv"
        )

        from pointcyto.data.metadata_regression import MetaDataRegression

        mymeta = gen_csv_meta(
            os.path.realpath(os.path.join(TESTS_DIR, pointcloud_toy_dataset_csvpheno)),
            path_phenodata="samples_pheno.csv",
            metadata_type=MetaDataRegression,
        )

        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        b = InMemoryPointCloud(metadata=mymeta)
        assert b.metadata == a.metadata
        assert torch.equal(b.pos, a.pos)
        assert a != b

        b.metadata.rm_root(True)
        # Now the dataset must be rebuilt
        c = InMemoryPointCloud(metadata=mymeta)
        assert c.metadata == a.metadata
        assert torch.equal(c.pos, a.pos)
        assert a != c
        repr(a)
        print(b)
        print(b.metadata)
        print(b._data)  # just for testing

    def test_InMemoryPointCloud_Reloading_with_str(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        b = InMemoryPointCloud(metadata=a.root)
        assert a != b
        assert b.metadata == a.metadata
        assert torch.equal(b.pos, a.pos)

        c = InMemoryPointCloud(metadata=pathlib.Path(a.root))
        assert a != c
        assert c.metadata == a.metadata
        assert torch.equal(c.pos, a.pos)

    def test_InMemoryPointCloud_item(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        # When a slice is selected, the returned value is again an InMemoryPointCloud
        # print(a[0:2])
        assert isinstance(a[0:2], InMemoryPointCloud)
        # If an int is selected, the returned value is a torch_geometric Data instance.
        # print(a[0])
        assert isinstance(a[0], tgeo.data.Data)

    def test_InMemoryPointCloud_subsetting(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        mymeta_train = mymeta[0:2, :]
        mymeta_val = mymeta[2:, :]
        a_train = InMemoryPointCloud(metadata=mymeta_train, clear_processed=True)
        a_val = InMemoryPointCloud(metadata=mymeta_val, clear_processed=True)
        assert len(a_train) == 2
        assert len(a_val) == 4

        complete_pointcloud = InMemoryPointCloud(metadata=mymeta)
        complete_pointcloud.process()
        pc_train = complete_pointcloud[0:2]
        pc_val = complete_pointcloud[2:]
        assert all(mymeta_train.pheno == pc_train.metadata.pheno)
        assert all(mymeta_val.pheno == pc_val.metadata.pheno)

        assert (
            mymeta_train.sample_feature_names == pc_train.metadata.sample_feature_names
        )
        assert mymeta_val.sample_feature_names == pc_val.metadata.sample_feature_names

    def test_InMemoryPointCloud_raw_dir_already_present(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw"),
            root=os.path.join(TESTS_DIR, pointcloud_toy_dataset),
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        print(a)

    def test_InMemoryPointCloud_sample_with_one_point_no_transform(self):
        mymeta = gen_foldering_meta(os.path.join(TESTS_DIR, empty_sample_dataset))
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        # print(a)
        for sample_x in a:
            # print(sample_x)
            assert len(sample_x.y) == 1
        b = InMemoryPointCloud(metadata=mymeta)
        for sample_x in b:
            # print(sample_x)
            assert len(sample_x.y) == 1

    def test_InMemoryPointCloud_sample_with_one_point_transform(self):
        # This is a pytorch-geometric/FixedPoints issue, see
        #   https://github.com/rusty1s/pytorch_geometric/issues/1090
        mymeta = gen_foldering_meta(os.path.join(TESTS_DIR, empty_sample_dataset))
        random_select_n_points = FixedPoints(num=100)
        a = InMemoryPointCloud(
            metadata=mymeta, clear_processed=True, transform=random_select_n_points
        )
        for sample_x in a:
            print(sample_x)
            assert len(sample_x.y) == 1

    def test_root_of_previous_test_issue(self):
        from torch_geometric.data.data import Data

        pointcloud = Data(pos=torch.randn(100, 3), y=torch.tensor([1]))
        single_point = Data(pos=torch.randn(1, 3), y=torch.tensor([1]))

        random_select_n_points = FixedPoints(num=10)

        print(random_select_n_points(pointcloud))
        # Data(pos=[10, 3], y=[1])
        assert len(random_select_n_points(pointcloud).y) == 1
        print(random_select_n_points(single_point))
        # Data(pos=[10, 3], y=[10])
        assert len(random_select_n_points(single_point).y) == 1

    def test_InMemoryPointCloud_different_raw_dirs_without_root(self):
        mymeta_toy = gen_foldering_meta(os.path.join(TESTS_DIR, pointcloud_toy_dataset))
        pointcloud_toy = InMemoryPointCloud(metadata=mymeta_toy, clear_processed=True)
        mymeta_tcell = gen_foldering_meta(os.path.join(TESTS_DIR, tcell_toy_dataset))
        pointcloud_tcell = InMemoryPointCloud(
            metadata=mymeta_tcell, clear_processed=True
        )
        assert (
            pointcloud_toy != pointcloud_tcell
        )  # the two pointclouds MUST not be the same.

        pointcloud_toy = InMemoryPointCloud(metadata=mymeta_toy)
        # Even after loading pointcloud_toy a second time, the data MUST NOT be the same
        if pointcloud_toy[0].pos.shape == pointcloud_tcell[0].pos.shape:
            assert not torch.allclose(pointcloud_toy[0].pos, pointcloud_tcell[0].pos)

    def test_InMemoryPointCloud_different_featurenames_in_data(self):
        mymeta = gen_foldering_meta(
            os.path.join(
                TESTS_DIR,
                pointcloud_toy_dataset,
                os.pardir,
                "Tcell_foldering_corrupted_colnames",
            ),
            root=os.path.join(TESTS_DIR, "new_datasets_dir"),
        )
        import copy

        saved_meta = copy.deepcopy(mymeta)
        assert saved_meta == mymeta
        with self.assertRaises(ValueError):
            my_pc = InMemoryPointCloud(metadata=mymeta)
            repr(my_pc)
        assert saved_meta == mymeta

    def test_InMemoryPointCloud_with_transform(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        random_select_n_points = FixedPoints(num=2)
        import numpy as np

        np.random.seed(123)
        a = InMemoryPointCloud(
            mymeta, transform=random_select_n_points, clear_processed=True
        )
        target_tensor = torch.tensor(
            [
                [
                    15.8737,
                    20.8384,
                    15.6227,
                    8.6955,
                    8.7534,
                    8.2956,
                    9.0078,
                    7.8170,
                    8.1263,
                    8.9850,
                    7.2125,
                    8.6033,
                    11.7536,
                    4107.0000,
                ],
                [
                    15.5161,
                    20.8384,
                    14.9606,
                    8.4246,
                    8.7429,
                    7.6892,
                    9.5300,
                    -5.6333,
                    11.7794,
                    8.1770,
                    6.9872,
                    8.1666,
                    12.4236,
                    2882.0000,
                ],
            ]
        )
        assert torch.allclose(a[0].pos, target_tensor, atol=1e-3, rtol=1e-3)

    def test_InMemoryPointCloud_FixedPoints_get_twice(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        random_select_n_points = FixedPoints(num=2)
        import numpy as np

        np.random.seed(123)
        a = InMemoryPointCloud(
            mymeta, transform=random_select_n_points, clear_processed=True
        )

        # Even the same a[N] called twice gives different results!!!
        # This is because inside, transform is called every time with a get() call.
        assert not torch.allclose(a[0].pos, a[0].pos, rtol=1e-4)

        # in contrast, when setting the seed for the RNG inside, the following results must be the same.
        # God, this is tricky.
        np.random.seed(123)
        tmp1 = a[0]
        np.random.seed(123)
        tmp2 = a[0]
        assert torch.allclose(tmp1.pos, tmp2.pos, rtol=1e-4)

    def test_InMemoryPointCloud_DataLoader(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        random_select_n_points = FixedPoints(num=2)
        import numpy as np

        np.random.seed(123)
        a = InMemoryPointCloud(
            mymeta, transform=random_select_n_points, clear_processed=True
        )

        # # The torch DataLoader cannot work anymore (2020-02-26) because I switched my PointCloud dataset
        # # completely to be based on torch_geometric.data.data.Data
        # t_loader = t_DataLoader(a)
        # for batch in t_loader:
        #     print(batch)
        #     break

        tg_loader = tg_DataLoader(a)
        for batch in tg_loader:
            repr(batch)
            print(batch)

    def test_InMemoryPointCloud_DataLoader_more_batches_than_samples(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        n_points_per_draw = 2
        random_select_n_points = FixedPoints(num=n_points_per_draw)
        import numpy as np

        np.random.seed(123)
        a = InMemoryPointCloud(
            mymeta, transform=random_select_n_points, clear_processed=True
        )

        # Default torch_geometric dataloader just selects the maximum possible amount of samples.
        tg_loader = tg_DataLoader(
            a, batch_size=len(mymeta) + 1
        )  # Draw more samples than in mymeta present
        for batch in tg_loader:
            assert batch.pos.shape[0] / len(batch.y) == n_points_per_draw
            assert len(batch.y) == len(mymeta)

        # torch_geometric.loader bases on torch.utils.data.DataLoader
        # therefore we can supply a "sampler"
        # The sampler defines the complete "sample-set" which can be drawn from during learning.
        # The batch_size splits this "sample-set" and iterating over the DataLoader then has
        #   num_samples / batch_size elements
        tg_loader = tg_DataLoader(
            a,
            batch_size=100,
            sampler=torch.utils.data.RandomSampler(
                data_source=a, replacement=True, num_samples=100
            ),
        )  # Draw more samples than in mymeta present
        for batch in tg_loader:
            assert batch.pos.shape[0] / len(batch.y) == n_points_per_draw
            assert len(batch.y) == 100

        tg_loader = tg_DataLoader(
            a,
            batch_size=20,
            sampler=torch.utils.data.RandomSampler(
                data_source=a, replacement=True, num_samples=100
            ),
        )  # Draw more samples than in mymeta present
        # 100 samples sampled randomly in total
        # 20 per batchsize
        # --> I have 5 batches
        assert len(tg_loader) == 5
        for batch in tg_loader:
            assert batch.pos.shape[0] / len(batch.y) == n_points_per_draw
            assert len(batch.y) == 20

        # What if the batch_size does not exactly part num_samples?
        tg_loader = tg_DataLoader(
            a,
            batch_size=20,
            sampler=torch.utils.data.RandomSampler(
                data_source=a, replacement=True, num_samples=90
            ),
        )  # Draw more samples than in mymeta present
        # 90 samples sampled randomly in total
        # 20 per batchsize
        # --> I have 5 batches and the last batch has just 10 samples
        assert len(tg_loader) == 5
        lenghts = []
        for batch in tg_loader:
            lenghts.append(len(batch.y))
            # assert batch.pos.shape[0] / len(batch.y) == n_points_per_draw
            # assert len(batch.y) == 20
        assert lenghts == [20, 20, 20, 20, 10]

    def test_InMemoryPointCloud_prepretransforms(self):
        from pointcyto.transforms.transform_param_onlist import NormalizePointCloudParam

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )

        mydataset = InMemoryPointCloud(mymeta, clear_processed=True)
        posmat: torch.Tensor = mydataset[0].pos
        # Without any normalization, the means are quite high.
        # 1 is just an empirical observation and has no further meaning
        assert torch.all(posmat.mean(dim=0).abs() > 1)

        mydataset = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=NormalizePointCloudParam(),
            clear_processed=True,
        )
        posmat: torch.Tensor = mydataset[0].pos
        # With normalization subtracting mean and dividing through standard deviation, the means are comparably.
        # 1 is just an empirical observation and has no further meaning
        assert torch.all(posmat.mean(dim=0).abs() < 1)

    def test_InMemoryPointCloud_reuse_prepretransforms(self):
        from pointcyto.transforms.transform_param_onlist import NormalizePointCloudParam

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        mydataset = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=NormalizePointCloudParam(),
            clear_processed=True,
        )

        # Ususally you would not reuse the same metadata!
        # But instead apply the pre_transform with parameters on a different metadata object.
        # But for testing this is enough.
        my_new_dataset_reusing_normalization_parameters = InMemoryPointCloud(
            mymeta,
            pre_transform=mydataset.pretransform_parameter[0]["related_transform"](
                **mydataset.pretransform_parameter[0]["param"]
            ),
            clear_processed=True,
        )
        posmat: torch.Tensor = my_new_dataset_reusing_normalization_parameters[0].pos
        # Even if I did not use the pre_pre_transform_param, the pre_transform also makes the normalization
        # But with given mean and standard deviation
        assert torch.all(posmat.mean(dim=0).abs() < 1)

    def test_InMemoryPointCloud_pre_pre_transform_param_list(self):
        from pointcyto.transforms.transform_param_onlist import NormalizePointCloudParam

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        mydataset = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=[
                NormalizePointCloudParam(),
                NormalizePointCloudParam(),
            ],
            clear_processed=True,
        )
        # The second normalization parameters are not zero
        # but much closer to zero than before
        # Probably the rest is just numerical instability
        assert torch.allclose(
            mydataset.pretransform_parameter[1]["param"]["mean"],
            torch.tensor(0, dtype=torch.float),
            atol=1e-2,
        )

    def test_InMemoryPointCloud_pre_pre_transform_param_None(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        mydataset = InMemoryPointCloud(
            mymeta, pre_pre_transform_param_onlist=None, clear_processed=True
        )
        print(mydataset)

    def test_InMemoryPointCloud_pre_pre_transform_param_list_when_reloading(self):
        from pointcyto.transforms.transform_param_onlist import NormalizePointCloudParam

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        mydataset = InMemoryPointCloud(
            mymeta,
            pre_pre_transform_param_onlist=NormalizePointCloudParam(),
            clear_processed=True,
        )
        with self.assertWarns(UserWarning):
            mydataset_reloaded = InMemoryPointCloud(
                mymeta,
                pre_pre_transform_param_onlist=NormalizePointCloudParam(),
                clear_processed=False,
            )
        assert torch.allclose(
            mydataset.pretransform_parameter[0]["param"]["mean"],
            mydataset_reloaded.pretransform_parameter[0]["param"]["mean"],
        )
        assert torch.allclose(
            mydataset.pretransform_parameter[0]["param"]["std"],
            mydataset_reloaded.pretransform_parameter[0]["param"]["std"],
        )
        assert torch.allclose(
            mydataset.pre_pre_transform_param_onlist[0].mean,
            mydataset_reloaded.pretransform_parameter[0]["param"]["mean"],
        )
        assert torch.allclose(
            mydataset.pre_pre_transform_param_onlist[0].std,
            mydataset_reloaded.pretransform_parameter[0]["param"]["std"],
        )

    def test_InMemoryPointCloud_change_metadata(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)

        # 1. Change only metadata.
        #       Problem: The PointCloud is already generated and does not get its information from the MetaData-object
        #                anymore. (It only uses it during .process())
        #                So even if the MetaData object and its "selection" is changed, that wont change anything for
        #                the processed PointCloud.
        # 1.1 Additional column
        df = a.metadata.pheno
        # adding a column to the phenodata
        # DO NOT DO IT LIKE THAT! (See error below and solution with MetaData.add_pheno_single())
        df.loc[:, "new_col"] = "a"
        print(a.metadata)

        # Selecting the new column works
        a.metadata.selection = "new_col"
        # BUT, accessing it NOT!
        with self.assertRaises(KeyError):
            # This cannot work as the MetaData object does not know about the new column.
            # class_name_id_map is missing
            print(a.metadata)
            # File "C:\Users\mi_so\PycharmProjects\ccc\ccc\datasets\metadata.py", line 245, in class_name_id_map
            #     return self._class_name_id_map[self.selection]
            # KeyError: 'D'
            #
            # That error occurs because it tries (but cannot) print the selected class levels.

        # To add a new column, do this:
        a.metadata.add_pheno_single(
            new_pheno_column=["a", "c", "c", "c", "e", "e"], col_name="new_col"
        )
        print(a.metadata)

        # 1.2 Changing existing data
        a.metadata.selection = "pheno_1"
        pre_selection_update_y = [str(single_pointcloud.y) for single_pointcloud in a]
        a.metadata.pheno.loc[:, "pheno_1"] = "X"
        a.metadata.selection = "pheno_1"
        a.metadata.update_selection()
        print(a.metadata.pheno)
        post_selection_update_y = [str(single_pointcloud.y) for single_pointcloud in a]

        # So here comes the problem:
        # The following holds TRUE, where we thought to have changed the metadata already!
        assert pre_selection_update_y == post_selection_update_y

    def test_InMemoryPointCloud_change_response(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        y_1 = [str(single_pointcloud.y) for single_pointcloud in a]
        # adding a column to the phenodata
        a.metadata.add_pheno_single(
            new_pheno_column=["a", "c", "c", "c", "c", "e"], col_name="new_col"
        )
        pre_selection_update_y = [str(single_pointcloud.y) for single_pointcloud in a]
        assert y_1 == pre_selection_update_y
        a.select_y_from_metadata("new_col")
        post_selection_update_y = [str(single_pointcloud.y) for single_pointcloud in a]
        assert pre_selection_update_y != post_selection_update_y

        a = InMemoryPointCloud(metadata=mymeta)
        pre_reloaded_selection_update_y = [
            str(single_pointcloud.y) for single_pointcloud in a
        ]
        assert pre_reloaded_selection_update_y == pre_selection_update_y

    def test_InMemoryPointCloud_split_pointcloud(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)

        assert len(a[1:3]) == 2
        data_part = [x.pos for x in a[1:3]]
        assert [x.shape for x in data_part] == [
            torch.Size([1000, 14]),
            torch.Size([1000, 14]),
        ]
        assert torch.allclose(data_part[0], a[1].pos)
        assert torch.allclose(data_part[1], a[2].pos)
        # mymeta.raw_filenames
        # ['A/P0I.fcs', 'A/PAB.fcs', 'A/PAB_copy.fcs', 'B/P0G.fcs', 'B/PAO.fcs', 'B/PAV.fcs']
        assert torch.allclose(data_part[0], data_part[1])
        assert len(a[1:4]) == 3

        data_part = [x.pos for x in a[1:4]]
        assert [x.shape for x in data_part] == [
            torch.Size([1000, 14]),
            torch.Size([1000, 14]),
            torch.Size([1000, 14]),
        ]
        assert torch.allclose(data_part[0], a[1].pos)
        assert torch.allclose(data_part[1], a[2].pos)
        assert torch.allclose(data_part[2], a[3].pos)
        # mymeta.raw_filenames
        # ['A/P0I.fcs', 'A/PAB.fcs', 'A/PAB_copy.fcs', 'B/P0G.fcs', 'B/PAO.fcs', 'B/PAV.fcs']
        assert torch.allclose(data_part[0], data_part[1])
        assert not torch.allclose(data_part[0], data_part[2])
        assert len(a[1:4]) == 3

        # NOTE: The complete data (_data) seems to stay the same always!
        # I hope that is no problem.
        assert a.pos.shape != a[1:3].pos.shape
        assert a._data.pos.shape == a[1:3]._data.pos.shape

    def test_InMemoryPointCloud_export_fcs(self):
        import flowio
        from numpy import reshape

        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        a.to_fcs("removeme")
        assert len(os.listdir("removeme")) == 2
        assert len(os.listdir("removeme/A")) == 3
        assert len(os.listdir("removeme/B")) == 3
        tmp = flowio.FlowData("removeme/B/PAO.fcs.fcs")
        assert tmp.text["p1r"] == "36"
        a.to_fcs("removeme_shifted", shift=100)
        npy_data_a = reshape(tmp.events, (-1, tmp.channel_count))
        assert len(os.listdir("removeme_shifted")) == 2
        assert len(os.listdir("removeme_shifted/A")) == 3
        assert len(os.listdir("removeme_shifted/B")) == 3
        tmp = flowio.FlowData("removeme_shifted/B/PAO.fcs.fcs")
        assert tmp.text["p1r"] != "36"
        npy_data_a_shifted = reshape(tmp.events, (-1, tmp.channel_count))

        assert npy_data_a.min() + 100 - npy_data_a_shifted.min() < 1e-4

    def test_InMemoryPointCloud_export_csv(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        with self.assertWarns(UserWarning):
            # UserWarning: In contrast to to_fcs(), only the pointcloud is saved, not the responses.
            a.to_csv("removeme")
        assert len(os.listdir("removeme")) == 2
        assert len(os.listdir("removeme/A")) == 3
        assert len(os.listdir("removeme/B")) == 3

    def test_InMemoryPointCloud_export_pt(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        a.to_pt("removeme")
        assert len(os.listdir("removeme")) == 2
        assert len(os.listdir("removeme/A")) == 3
        assert len(os.listdir("removeme/B")) == 3

    def test_InMemoryPointCloud_export_xxx(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        for xxx in ["jay", "fcs", "csv", "pt"]:
            a._to_xxx("removeme", filetype=xxx)
            assert len(os.listdir("removeme")) == 2
            assert len(os.listdir("removeme/A")) == 3
            assert len(os.listdir("removeme/B")) == 3
            shutil.rmtree("removeme")

    def test_benchmark_InMemoryPointCloud_export_xxx(self):
        do_large_benchmark = False  # Done 2021-01-19
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        if do_large_benchmark:
            benchmark_size = 7  # Done 2021-01-19
            for i in range(benchmark_size):
                # for i in range(2):
                # 0, 1, 2, 3, 4, 5,
                # 6, 12, 24, 48, 96, 192, xxx, 768
                mymeta = mymeta.concat(mymeta, ignore_identical_filenames=True)

        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)
        # to have unique files:
        tmp = [f"file_{i}.fcs" for i in range(len(a))]
        a.metadata.raw_filenames = tmp

        import time

        times = {}
        dir_sizes = {}
        difftimes = {"write": {}, "read_meta": {}, "read_data": {}}
        # all_types = ["pt", "jay", "fcs", "csv", "feather"]
        # all_types = [
        #     f.replace(".py", "")
        #     for f in os.listdir("src/pointcyto/io/loaders")
        #     if f.endswith(".py") and f != "__init__.py"
        # ]
        from pointcyto.io.parse_by_ext import POINTCLOUD_EXTENSIONS as all_types_dot

        all_types = [x.replace(".", "") for x in all_types_dot]

        print("\n ------- Start saving -------\n")
        for xxx in all_types:
            times[xxx] = {}
            times[xxx]["start_write"] = time.time()
            a._to_xxx(f"removeme_{xxx}", filetype=xxx, verbose=True)
            times[xxx]["end_write"] = time.time()
            timediff_write = times[xxx]["end_write"] - times[xxx]["start_write"]
            print(f"{xxx:10}: {timediff_write}")
            difftimes["write"][xxx] = timediff_write
            dir_sizes[xxx] = sum(
                os.path.getsize(os.path.join(f"removeme_{xxx}", f)) / 1048576
                for f in os.listdir(f"removeme_{xxx}")
            )

        print("\n ------- Start reading -------\n")
        for xxx in all_types:
            times[xxx] = {}
            times[xxx]["start_read_meta"] = time.time()
            meta_removeme = gen_meta_filenames(f"removeme_{xxx}")
            times[xxx]["start_read_data"] = time.time()
            impc_removeme = InMemoryPointCloud(
                metadata=meta_removeme, clear_processed=True
            )
            times[xxx]["end_read_data"] = time.time()
            print(impc_removeme)
            d_read_meta = times[xxx]["start_read_data"] - times[xxx]["start_read_meta"]
            d_read_data = times[xxx]["end_read_data"] - times[xxx]["start_read_data"]
            print(f"Read meta {xxx:10}: {d_read_meta}")
            print(f"Read data {xxx:10}: {d_read_data}\n")
            difftimes["read_meta"][xxx] = d_read_meta
            difftimes["read_data"][xxx] = d_read_data
        print("\n ------- End reading -------\n")

        for key, value in difftimes.items():
            print(key)
            for k, v in value.items():
                print("   ", k, v)
        print(f"Directory sizes for {len(meta_removeme)} files")
        for key, value in dir_sizes.items():
            print("    ", key, round(value, 2), "MB")
        # Test 2023-01-19, rhskl1
        # write
        #     pt 4.566260099411011
        #     jay 27.967937707901
        #     fcs 59.25578022003174
        #     csv 16.594340324401855
        # read_meta
        #     pt 0.008815765380859375
        #     jay 0.007297039031982422
        #     fcs 0.0071790218353271484
        #     csv 0.007416725158691406
        # read_data
        #     pt 8.280493259429932
        #     jay 35.553593158721924
        #     fcs 15.747443675994873
        #     csv 51.12806057929993
        # Directory sizes for 768 files
        #      pt 5471.66 MB
        #      jay 5471.54 MB
        #      fcs 5471.8 MB
        #      csv 12922.16 MB

        for xxx in all_types:
            shutil.rmtree(f"removeme_{xxx}")

    def test_InMemoryPointCloud_different_featurenames(self):
        mymeta = gen_foldering_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        )
        a = InMemoryPointCloud(metadata=mymeta, clear_processed=True)

        with self.assertWarns(UserWarning):
            # UserWarning: In contrast to to_fcs(), only the pointcloud is saved, not the responses.
            a.to_csv("removeme")

        assert len(os.listdir("removeme")) == 2
        assert len(os.listdir("removeme/A")) == 3
        assert len(os.listdir("removeme/B")) == 3

        import pandas as pd

        single_sample = pd.read_csv("removeme/A/PAB.fcs.csv")
        single_sample.columns = ["new_" + x for x in single_sample.columns]
        single_sample.to_csv("removeme/A/PAB.fcs.csv", index=False)

        mymeta_different_featurenames = gen_foldering_meta("removeme")

        with self.assertRaises(ValueError):
            # The following error must occur:
            #   ValueError: Unequal feature names of
            #   A/PAB.fcs.csv:
            a_different_featurenames = InMemoryPointCloud(
                metadata=mymeta_different_featurenames, clear_processed=True
            )
            print(a_different_featurenames)

        mymeta_different_featurenames_ignore = gen_foldering_meta(
            "removeme", ignore_unequal_feature_names=True
        )
        # catch all warnings, there must not be any userwarnings
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            a_different_featurenames_ignore = InMemoryPointCloud(
                metadata=mymeta_different_featurenames_ignore, clear_processed=True
            )
        a_different_featurenames_ignore = InMemoryPointCloud(
            metadata=mymeta_different_featurenames_ignore,
            clear_processed=True,
        )
        assert (
            a_different_featurenames_ignore.metadata.feature_names[0]
            == "Feature names NOT identical over samples!"
        )

        with self.assertWarns(UserWarning):
            # UserWarning: In contrast to to_fcs(), only the pointcloud is saved, not the responses.
            a_different_featurenames_ignore.to_csv("removeme_2")

        single_sample = pd.read_csv("removeme_2/A/PAB.fcs.csv.csv")
        assert all([x.startswith("new_") for x in single_sample.columns])
        single_sample = pd.read_csv("removeme_2/A/P0I.fcs.csv.csv")
        assert all([not x.startswith("new_") for x in single_sample.columns])
        assert len(os.listdir("removeme_2")) == 2
        assert len(os.listdir("removeme_2/A")) == 3
        assert len(os.listdir("removeme_2/B")) == 3

    def tearDown(self) -> None:
        try:
            shutil.rmtree(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "processed"))
            pass
        except FileNotFoundError:
            pass
        try:
            shutil.rmtree("removeme")
            pass
        except FileNotFoundError:
            pass
        try:
            shutil.rmtree(
                os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw", "processed")
            )
            shutil.rmtree(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw", "raw"))
            pass
        except FileNotFoundError:
            pass
        try:
            shutil.rmtree(os.path.join(TESTS_DIR, "new_datasets_dir"))
        except FileNotFoundError:
            pass



import torch


def randompointcloud(n_points: int = 4, n_features: int = 5):
    """
    Generate a random torch point cloud

    Args:
        n_points:
            Number of points
        n_features:
            Number of features

    Returns:
        a random torch point cloud with dimension (n_points, n_features)
    """
    return torch.rand((n_points, n_features))



import torch

# See also https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3243046/


def arcsinh(x: torch.Tensor) -> torch.Tensor:
    return torch.asinh(x)


def arcsinh_param(x: torch.Tensor, a, b, c) -> torch.Tensor:
    """

    Args:
        x: Data (torch.Tensor)
        a: positive float, shifting about 0 inside arcsinh
        b: positive float, scale factor
        c: float, shifting the arcsinh result

    Returns:
        ..math::
            f(x, a, b, c) = arcsinh(a + bx) + c
    """
    return arcsinh(a + b * x) + c


def scale_low_high_01(tensor, low_high=None, inplace=False):
    if not inplace:
        tensor = tensor.clone()

    dtype = tensor.dtype
    if len(low_high) == 1:
        # generate the low_high_tensor for all features (=columns)
        low_high = low_high * tensor.shape[1]
    # With the following line I ensure that there are _always_ low AND high values.
    # If there is only one value (len(x) == 1), then
    #   low -> x
    #   high -> low + 1
    # which results in
    #   (tensor - low) / (high - low) =
    #   (tensor - low) / (low + 1 - low) =
    #   (tensor - low) / (1) = tensor - low

    low_high_lowhigh = [[x[0], x[0] + 1] if len(x) == 1 else x for x in low_high]
    if not all([len(x) == 2 for x in low_high_lowhigh]):
        raise ValueError("There are not always exactly 2 values, one low and one high.")
    low_high_tensor = torch.as_tensor(
        low_high_lowhigh, dtype=dtype, device=tensor.device
    )
    if len(low_high_tensor) != tensor.shape[1]:
        raise ValueError(
            "Neither only one, nor a matching number of low_high "
            + f"scalings{len(low_high_tensor)} to the number of features.{tensor.shape[1]}"
        )

    # See QminmaxPointCloud for the transpose
    low_high_tensor = low_high_tensor.t()
    return (tensor - low_high_tensor[0, :]) / (
        low_high_tensor[1, :] - low_high_tensor[0, :]
    )


def scale_featurewise(tensor, feature_scaling_vector=None, inplace=False):
    """
    Divide the tensor per column by the corresponding value in feature_scaling_vector
    """
    if not inplace:
        tensor = tensor.clone()

    dtype = tensor.dtype
    if len(feature_scaling_vector) == 1:
        # generate the low_high_tensor for all features (=columns)
        feature_scaling_vector = feature_scaling_vector * tensor.shape[1]
    feature_scaling_vector_tensor = torch.as_tensor(
        feature_scaling_vector, dtype=dtype, device=tensor.device
    )
    if len(feature_scaling_vector_tensor) != tensor.shape[1]:
        raise ValueError(
            "Neither only one, nor a matching number of scalings "
            + f"{len(feature_scaling_vector_tensor)} to the number of features.{tensor.shape[1]}"
        )
    tensor.div_(feature_scaling_vector_tensor)
    return tensor



from unittest import TestCase

import torch

import pointcyto.transforms.transforms_functional as t_func


class TestTransformsFunctional(TestCase):
    def test_asinh(self):
        torch.manual_seed(23178)
        x = torch.randn((3, 2))
        target = torch.tensor(
            [
                [-1.02025531310114, 0.650699301054285],
                [0.156954780747593, 0.646511502753712],
                [0.896005194544506, -0.591923641697527],
            ]
        )
        # target via R:
        # x = matrix(c(-1.2067, 0.6976,
        #              0.1576, 0.6925,
        #              1.0208, -0.6271), ncol=2, byrow=TRUE)
        # calced = asinh(x)
        # Rvarious::print_as_python(calced)
        assert torch.allclose(t_func.arcsinh(x), target, atol=1e-4)

    def test_asinh_param(self):
        torch.manual_seed(23178)
        x = torch.randn((3, 2))

        # target via R:
        # x = matrix(c(-1.2067, 0.6976,
        #              0.1576, 0.6925,
        #              1.0208, -0.6271), ncol=2, byrow=TRUE)
        # a = 1
        # b = 1
        # c = 2
        # calced = asinh(a + b * x) + c
        # Rvarious::print_as_python(calced)
        target = torch.tensor(
            [
                [1.7947442733302, 3.29960294002619],
                [2.98854374326608, 3.29701153276323],
                [3.45289900843958, 2.3647576349827],
            ]
        )
        assert torch.allclose(t_func.arcsinh_param(x, 1, 1, 2), target, atol=1e-4)



# These transformations follow the nomenclature from flowCore
# https://bioconductor.org/packages/devel/bioc/vignettes/flowCore/inst/doc/HowTo-flowCore.pdf
# Section 6.1 "Standard transformations
from typing import List

from torch_geometric.data.dataset import Dataset

from pointcyto.transforms.transforms import functional_transform_adapter
from pointcyto.transforms.transforms_functional import arcsinh_param
from pointcyto.transforms.utils import standard_repr


class TransformFunctional(object):
    def __init__(
        self, original_featurenames: List[str] = None, dataset_attribute: str = "pos"
    ):
        self.attr = dataset_attribute
        self.orig_featurenames = original_featurenames

    def __call__(self, data: Dataset):
        # setattr(data, self.attr,
        #         functional_transform_adapter(
        #             # these parameters must always be present for functional_transform_adapter()
        #             data, <the_function>, self.attr,
        #             # These are the actual arguments to the function
        #             self.feature_indices, self.inplace))
        # return data
        raise NotImplementedError


class TransformFunctionalListadapter(object):
    def __init__(
        self, original_featurenames: List[str] = None, dataset_attribute: str = "pos"
    ):
        self.attr = dataset_attribute
        self.orig_featurenames = original_featurenames

    def __call__(self, data_list: List[Dataset]):
        # for sample in data_list:
        #     setattr(sample, self.attr,
        #             functional_transform_adapter(
        #                 # these parameters must always be present for functional_transform_adapter()
        #                 data, <the_function>, self.attr,
        #                 # These are the actual arguments to the function
        #                 self.feature_indices, self.inplace))
        # return data_list, {'return_parameter1': self.mean, 'return_parameter2': self.std}
        raise NotImplementedError


class ArcsinhTransform(TransformFunctional):
    def __init__(
        self,
        original_featurenames: List[str] = None,
        dataset_attribute: str = "pos",
        a: float = 1,
        b: float = 1,
        c: float = 0,
    ):
        super(ArcsinhTransform, self).__init__(
            original_featurenames=original_featurenames,
            dataset_attribute=dataset_attribute,
        )
        self.a = a
        self.b = b
        self.c = c

    def __call__(self, data: Dataset):
        setattr(
            data,
            self.attr,
            functional_transform_adapter(
                # these parameters must always be present for functional_transform_adapter()
                data,
                arcsinh_param,
                self.attr,
                # These are the actual arguments to the function
                a=self.a,
                b=self.b,
                c=self.c,
            ),
        )
        return data

    def __repr__(self):
        return standard_repr(self)


class ArcsinhTransformList(ArcsinhTransform):
    def __call__(self, data_list: List[Dataset]):
        for sample in data_list:
            setattr(
                sample,
                self.attr,
                functional_transform_adapter(
                    # these parameters must always be present for functional_transform_adapter()
                    sample,
                    arcsinh_param,
                    self.attr,
                    # These are the actual arguments to the function
                    a=self.a,
                    b=self.b,
                    c=self.c,
                ),
            )
        return data_list, {}



import os
import pathlib
import pickle
import time
import warnings
from math import inf
from typing import Any, List, Optional

import pytorch_lightning
import torch_geometric
from torch_geometric.data.dataset import Dataset

from pointcyto.data.InMemoryPointCloud import InMemoryPointCloud
from pointcyto.data.metadata_base import MetaDataBase


def generate_load_data_wait(
    metadata: MetaDataBase,
    lock_file: str = "/tmp/lock.generate_load_data",
    wait_seconds: int = 5,
    max_wait_seconds: int = None,
    **kwargs
):
    def lock_process_data():
        # generate a lock
        open(lock_file, "a").close()
        # generate the data
        res = generate_load_data(metadata=metadata, **kwargs)
        # remove the lock after finishing
        os.remove(lock_file)
        return res

    if max_wait_seconds is None:
        max_wait_seconds = inf  # from math
    t0 = time.time()
    if os.path.exists(lock_file):
        while time.time() - t0 < max_wait_seconds:
            if os.path.exists(lock_file):
                print("Waiting because ", lock_file, " exists")
                time.sleep(wait_seconds)
            else:
                # when the lock_file does not exist anymore we can load the data
                res = generate_load_data(metadata=metadata, **kwargs)
                # After loading we can exit the waiting-loop.
                break
        if os.path.exists(lock_file):
            # If the file STILL exists after the waiting time, remove it by force, maybe the
            # generating process failed?
            print(
                "Removing ",
                lock_file,
                " EVEN IF IT EXISTS after max_wait_seconds was exceeded!",
            )
            os.remove(lock_file)
            res = lock_process_data()
    else:
        res = lock_process_data()

    return res


def generate_load_data(
    metadata: MetaDataBase,
    datagenerator_seed: int = 51284,
    basis_pointcloud: InMemoryPointCloud = None,
    prepretrans_parameters_picklefile: pathlib.Path = None,
    pointcloud_generator: Dataset = InMemoryPointCloud,
    transform: Optional[Any] = None,
    pre_transform: List[Any] = None,
    pre_pre_transform_param_onlist=None,
    clear_processed: bool = False,
    **kwargs
):
    """_summary_

    Args:
        metadata (MetaDataBase): _description_
        datagenerator_seed (int): _description_
        basis_pointcloud (InMemoryPointCloud, optional): _description_. Defaults to None.
        pointcloud_generator (Dataset, optional): _description_. Defaults to InMemoryPointCloud.
        transform (Optional[Any], optional): _description_. Defaults to None. E.g.
            FixedPoints(
                num=n_points_per_sample, replace=False, allow_duplicates=False
            )
        pre_transform (List[Any], optional): _description_. Defaults to None.
        pre_pre_transform_param_onlist (_type_, optional): _description_. Defaults to None.
        clear_processed (bool, optional): _description_. Defaults to False.
        kwargs:
            Further arguments to pointcloud_generator

    Returns:
        _type_: _description_
    """
    if basis_pointcloud is not None or prepretrans_parameters_picklefile is not None:
        if prepretrans_parameters_picklefile is None:
            # If there is a basis pointcloud (usually the training data) which should be used, load their
            # pretransform_parameters.
            prepretrans_parameters_picklefile = os.path.join(
                basis_pointcloud.processed_dir, "pretransform_parameter.pickle"
            )
        else:
            warnings.warn(
                "You gave basis_pointcloud and prepretrans_parameters_picklefile\n"
                + "Only the prepretrans_parameters_picklefile is used."
            )

        # Use the saved pretransform parameters which are saved by pre_pre_transform_param_onlist
        # to use then as pre_transform for the validation dataset
        with open(prepretrans_parameters_picklefile, "rb") as f:
            pre_pre_transforms_saved = pickle.load(f)

        insert_ppt = [
            single_pretransform["related_transform"](**single_pretransform["param"])
            for single_pretransform in pre_pre_transforms_saved
        ]
        if pre_transform is not None:
            if not isinstance(pre_transform, list):
                pre_transform = [pre_transform]
            pre_transform = pre_transform + insert_ppt

        else:
            pre_transform = insert_ppt
        pre_transform_composed = torch_geometric.transforms.compose.Compose(
            pre_transform
        )
        if pre_pre_transform_param_onlist is not None:
            warnings.warn(
                "Are you sure you want to add a pre_pre_transform_param_onlist when supplying a basis_pointcloud?"
            )
    else:
        pre_transform_composed = pre_transform

    pytorch_lightning.seed_everything(datagenerator_seed)
    generated_pointcloud = pointcloud_generator(
        metadata=metadata,
        transform=transform,
        pre_transform=pre_transform_composed,
        pre_pre_transform_param_onlist=pre_pre_transform_param_onlist,
        clear_processed=clear_processed,
        **kwargs
    )
    return generated_pointcloud



import os
import pickle
import shutil
import warnings
from pathlib import Path
from typing import Union

import datatable as dt
import flowio
import torch
from numpy.random import seed
from torch_geometric.data import Data, InMemoryDataset
from torch_geometric.data.dataset import _repr as tg_function_repr

from pointcyto.data.metadata_base import MetaDataBase
from pointcyto.io.parse_by_ext import parse_by_ext

try:
    import pyarrow.feather as feather

    pyarrow_installed = True
except ImportError:
    pyarrow_installed = False
    pass


class InMemoryPointCloud(InMemoryDataset):
    """
    For many examples check out the test suites: :class:`tests.datasets.test_InMemoryPointCloud.TestInMemoryPointCloud`
    """

    def __init__(
        self,
        metadata: Union[MetaDataBase, Path, str],
        transform=None,
        pre_pre_transform_param_onlist=None,
        pre_transform=None,
        clear_processed: bool = False,
        loading_function=parse_by_ext,
        loading_function_arguments=None,
    ):
        """
        Generates an InMemoryPointCloud based on the files inside metadata.

        In this root-directory, if not present, "raw/" and "processed/" directories are created.
        In "raw/", the original data can but does not have to lie.
        In "processed/", the processed data is saved.

        Args:
            metadata:
                A MetaData object containing all relevant information
            transform:
                Transformations for the pytorch_geometric Data object
            pre_pre_transform_param_onlist:
                Transformations for a list of pytorch_geometric Data objects.
                Must be a transformation on the dataset and returns a tuple:
                ::

                    data_list, {'name': self.name, 'related_transform': self.related_transform,
                                'param': {'mean': self.mean, 'std': self.std}}

                This tuple holds the transformed data_list and a dictionary for further use.
                The dictionary includes the name of the transformation, the related **not onlist** transformation and
                the parameters which are a result from the onlist-transformation to be used in the
                related transformation. The standard way to use the related transformation is to loop over
                ``self.pretransform_parameter`` as shown in
                 :meth:`tests/transforms/test_pre_pre_transforms_onlist.TestPrePreTransformsOnList.test_NormalizePointCloudParam_pre_pre_transform_param_list_save_load`
                (Look at the part with the ``recreated_transform``)


                ``pre_pre_transforms`` are applied *before* ``pre_transform``.

                Examples:
                    - :class:`pointcyto.transforms.transform_param_onlist.NormalizePointCloudParam`
                    - :class:`tests.transforms.test_pre_pre_transforms_onlist.TestPrePreTransformsOnList`

            pre_transform:
                Transformations for the pytorch_geometric Data object
            loading_function:
                function used in process() to load the data.
                Should return a dictionary with the following keys
                - ``features``: A list with names of the features
                - ``point_data``: A torch.Tensor with the actual loaded data.

            loading_function_arguments:
                Arguments additionally supplied to function used in process() to load the data.
                The function call is:

                .. code-block:: python

                    parsed = self.loading_function(
                        path=os.path.join(self.metadata.orig_dir, metadata_X["raw_filenames"]),
                        y=metadata_X["y"],
                        **self.loading_function_args
                    )

        Returns:
            No return
        """
        self.loading_function = loading_function
        if loading_function_arguments is None:
            self.loading_function_args = {}
        else:
            self.loading_function_args = loading_function_arguments

        self.y_from_metadata_selection: str = None

        if clear_processed:
            try:
                shutil.rmtree(os.path.join(metadata.root, "processed"))
            except FileNotFoundError:  # then it was already deleted
                pass

        self.warnings_captured = {}
        self.loaded_or_processed = "loaded"
        if isinstance(metadata, str) or isinstance(metadata, Path):
            if os.path.isdir(metadata):
                # then it might be the given autoroot, otherwise will fail
                metadata_file = os.path.join(
                    metadata, "processed", "pointcloud_metadata.pickle"
                )
            else:
                metadata_file = metadata
                # Then it should be an actual <somename_pickled_MetaData_object>.pickle
            with open(os.path.join(metadata_file), "rb") as f:
                self.metadata = pickle.load(f)
            # A given, pickled metadata FILE must always be in the folder
            #   rootdir/processed/pointcloud_metadata.pickle
            # Therefore, the new root is  rootdir/processed/.. -> rootdir
            self.metadata.root = os.path.realpath(
                os.path.join(os.path.dirname(metadata_file), os.pardir)
            )
        else:
            self.metadata = metadata

        # self._reloading_processed_dir = False
        # if isinstance(metadata, str) or isinstance(metadata, Path):
        #     if os.path.isdir(metadata):
        #         # then it might be the given autoroot, otherwise will fail
        #         metadata_file = os.path.join(
        #             metadata, "processed", "pointcloud_metadata.pickle"
        #         )
        #     else:
        #         metadata_file = metadata
        #         # Then it should be an actual <somename_pickled_MetaData_object>.pickle
        #     with open(os.path.join(metadata_file), "rb") as f:
        #         self.metadata = pickle.load(f)

        #     # A given, pickled metadata FILE must always be in the folder
        #     #   rootdir/processed/pointcloud_metadata.pickle
        #     # Therefore, the new root is  rootdir/processed/.. -> rootdir
        #     self.metadata.root = os.path.realpath(
        #         os.path.join(os.path.dirname(metadata_file), os.pardir)
        #     )
        #     self._reloading_processed_dir = True
        # else:
        #     self.metadata = metadata

        if (
            isinstance(pre_pre_transform_param_onlist, list)
            or pre_pre_transform_param_onlist is None
        ):
            self.pre_pre_transform_param_onlist = pre_pre_transform_param_onlist
        else:
            self.pre_pre_transform_param_onlist = [pre_pre_transform_param_onlist]
        self.pretransform_parameter = list()

        # process is called inside here, therefore everything which is used in process must be defined before.
        super(InMemoryPointCloud, self).__init__(
            root=self.metadata.root, transform=transform, pre_transform=pre_transform
        )
        # # delattr(self, "raw_dir")
        # del self.raw_dir
        # self.raw_dir = self.metadata.orig_dir  # origdir might be something different than the root/raw-directory.
        self.data, self.slices = torch.load(
            os.path.join(self.processed_dir, self.processed_file_names[0])
        )
        for warning_location in self.warnings_captured:
            if (
                warning_location == "_process_userwarnings"
                and self.loaded_or_processed == "loaded"
            ):
                continue
            else:
                for warning_i in self.warnings_captured[warning_location]:
                    warnings.warn(warning_i, category=UserWarning)

    # Overwrite the raw_dir property from DataSet because
    # origdir might be something different than the "root/raw/" -directory.
    @property
    def raw_dir(self):
        return self.metadata.orig_dir

    @property
    def raw_file_names(self):
        # A list of files in the raw_dir which needs to be found in order to skip the download.
        return self.metadata.raw_filenames

    @property
    def processed_file_names(self):
        # A list of files in the processed_dir which needs to be found in order to skip the processing.
        return ["pointcloud_inmemory.pt"]

    def _to_xxx(
        self,
        outdir=".",
        verbose: bool = True,
        filetype="csv",
        shift=0,
        manual_seed=213124876,
    ):
        if manual_seed is not None:
            seed(manual_seed)  # np.random.seed(manual_seed)
            torch.manual_seed(manual_seed)
        for sample_i, sample_x in enumerate(self):
            if manual_seed is not None:
                seed(manual_seed)  # np.random.seed(manual_seed)
                torch.manual_seed(manual_seed)
            if verbose:
                print(sample_x)
            featurenames = self.metadata.sample_feature_names[sample_i]
            if featurenames is None:
                featurenames = ["f_" + str(i) for i in range(sample_x.pos.shape[1])]
            # Similar to
            #   /home/gugl/.conda_envs/ccc_optuna_III/lib/python3.8/site-packages/flowkit/_models/sample.py
            current_fp = os.path.join(
                outdir, self.raw_file_names[sample_i] + "." + filetype
            )
            os.makedirs(os.path.dirname(current_fp), exist_ok=True)
            if shift != 0:
                # FCS files make problems in Kaluza if they have values below 0.
                # Therefore we introduce a way to add shifts to the data.
                # All cells get the exact same shift
                sample_x.pos += torch.tensor(shift)
            if filetype == "pt":
                # Have to clone because subsetting does not actually create a new tensor
                # https://discuss.pytorch.org/t/saving-tensor-with-torch-save-uses-too-much-memory/46865/3
                torch.save(obj=[sample_x.pos.clone(), featurenames], f=current_fp)
            else:
                pos_numpy = sample_x.pos.numpy()
                pos_numpy_max_ceiling = sample_x.pos.max(axis=0)[0].ceil().int()
                if filetype == "fcs":
                    metadata_dict = {
                        f"response_element_{i}": str(element_i)
                        for i, element_i in enumerate(
                            sample_x.y.numpy().flatten().tolist()
                        )
                    }
                    # Update pNr for "parameter N range"
                    metadata_dict.update(
                        {
                            f"p{feature_i}r": str(
                                int(pos_numpy_max_ceiling[feature_i] * 1.2 + 10)
                            )
                            for feature_i in range(sample_x.pos.shape[1])
                        }
                    )
                    with open(current_fp, "wb") as f:
                        flowio.create_fcs(
                            event_data=pos_numpy.flatten().tolist(),
                            metadata_dict=metadata_dict,
                            channel_names=featurenames,
                            file_handle=f,
                        )
                elif filetype == "csv":
                    pos_dt = dt.Frame(pos_numpy, names=featurenames)
                    pos_dt.to_csv(current_fp)
                elif filetype == "jay":
                    pos_dt = dt.Frame(pos_numpy, names=featurenames)
                    pos_dt.to_jay(current_fp)
                elif filetype == "feather":
                    if not pyarrow_installed:
                        raise ImportError(
                            "Did not find pyarrow, cannot save feather files"
                        )
                    pos_dt = dt.Frame(pos_numpy, names=featurenames)
                    feather.write_feather(
                        df=feather.Table.from_pandas(pos_dt.to_pandas()),
                        dest=current_fp,
                    )

            if verbose:
                print("Saved ", current_fp, "\n")

    def to_csv(self, outdir=".", verbose: bool = True):
        warnings.warn(
            "In contrast to to_fcs(), only the pointcloud is saved, not the responses."
        )
        self._to_xxx(outdir=outdir, verbose=verbose, filetype="csv")

    def to_fcs(self, outdir=".", verbose: bool = True, shift: bool = 0):
        self._to_xxx(outdir=outdir, verbose=verbose, filetype="fcs", shift=shift)

    def to_pt(self, outdir=".", verbose: bool = True):
        self._to_xxx(outdir=outdir, verbose=verbose, filetype="pt")

    def to_feather(self, outdir=".", verbose: bool = True):
        self._to_xxx(outdir=outdir, verbose=verbose, filetype="feather")

    def to_jay(self, outdir=".", verbose: bool = True):
        self._to_xxx(outdir=outdir, verbose=verbose, filetype="jay")

    def download(self):
        """
        I will never download data as they are assumed to be local.

        But here you could  download raw data into raw_dir.

        """
        pass

    def _process(self):
        f = os.path.join(self.processed_dir, "str_pre_pre_transform_param.pickle")
        self.warnings_captured.update({"pre_pre_transform_param": []})
        if os.path.exists(f):
            # if os.path.exists(f):
            with open(f, "rb") as myfile:
                previous_param_as_str = pickle.load(myfile)
            if previous_param_as_str != tg_function_repr(
                self.pre_pre_transform_param_onlist
            ):
                warnstring = (
                    "The `pre_pre_transform_param` argument differs from the one used in "
                    + "the pre-processed version of this dataset. If you really "
                    + "want to make use of another pre-processing technique, make "
                    + "sure to delete `{}` first.".format(self.processed_dir)
                )
                # warnings.warn(warnstring)
                self.warnings_captured.update({"pre_pre_transform_param": warnstring})
        try:
            with open(
                os.path.join(self.processed_dir, "pretransform_parameter.pickle"), "rb"
            ) as f:
                self.pretransform_parameter = pickle.load(f)
        except FileNotFoundError:
            pass  # then the pre_pre_transform_param.pt just does not exist

        _process_userwarnings = []
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            super(InMemoryPointCloud, self)._process()
        for warning_x in w:
            if isinstance(warning_x.category(), UserWarning):
                _process_userwarnings += [warning_x]
            else:
                warnings.warn(warning_x, category=warning_x.category)
        self.warnings_captured.update({"_process_userwarnings": _process_userwarnings})

    def process(self):
        """
        Processes raw data and saves it into the processed_dir.
        Reads data into huge `Data` list.

        It
         - iterates over all ``raw_filenames`` from self.metadata
         - reads(parses) the file
         - Extracts the feature names and adds them to the respective metadata sample
         - Applies ``pre_filter``
         - Applies ``pre_pre_transform_param_onlist``
         - Saves the parameters and transformations from ``pre_pre_transform_param_onlist``
         into ``self.processed_dir/'pretransform_parameter.pickle'``
         - Applies ``pre_transform``s
         - Collates (combines) and saved the final processed dataset.

        Returns:
            None
        """
        self.loaded_or_processed = "processed"
        sample_feature_names = []
        data_list = []
        self.y_from_metadata_selection = self.metadata.selection
        for metadata_N, metadata_X in enumerate(self.metadata):
            print(
                str(metadata_N + 1),
                "/",
                str(len(self.metadata)),
                "  ",
                os.path.join(self.metadata.orig_dir, metadata_X["raw_filenames"]),
            )
            # Read data from `raw_path`.
            # now the data are in the format of tuple:(Tensor, markernames)
            # locals()["myfunction"]()
            current_response = metadata_X["y"]
            parsed = self.loading_function(
                path=os.path.join(self.metadata.orig_dir, metadata_X["raw_filenames"]),
                y=current_response,
                **self.loading_function_args,
            )

            sample_feature_names += [parsed["features"]]
            # hasattr to be backwards compatible.
            if (
                not hasattr(self.metadata, "ignore_unequal_feature_names")
                or not self.metadata.ignore_unequal_feature_names
            ):
                if sample_feature_names[0] != parsed["features"]:
                    raise ValueError(
                        "Unequal feature names of \n"
                        + "".join([self.metadata.raw_filenames[metadata_N]])
                        + ":"
                        + "\n  featurenames: "
                        + ", ".join(parsed["features"])
                        + "\n  sample 0:     "
                        + ", ".join(sample_feature_names[0])
                    )
            data_list += [parsed["point_data"]]
        self.metadata.sample_feature_names = sample_feature_names
        print("pre_filter")
        if self.pre_filter is not None:
            data_list = [data for data in data_list if self.pre_filter(data)]

        print("pre_pre_transform_param_onlist")
        if self.pre_pre_transform_param_onlist is not None:
            for transformation in self.pre_pre_transform_param_onlist:
                data_list, param_list = transformation(data_list)
                self.pretransform_parameter.append(param_list)

        print("pre_pre_transform_param_onlist saving")
        with open(
            os.path.join(self.processed_dir, "str_pre_pre_transform_param.pickle"), "wb"
        ) as f:
            pickle.dump(tg_function_repr(self.pre_pre_transform_param_onlist), f)
        with open(
            os.path.join(self.processed_dir, "pretransform_parameter.pickle"), "wb"
        ) as f:
            pickle.dump(self.pretransform_parameter, f)

        print("pre_transform")
        if self.pre_transform is not None:
            data_list = [self.pre_transform(data) for data in data_list]

        print("collate")
        data, slices = self.collate(data_list)

        print("saving")
        torch.save((data, slices), self.processed_paths[0])
        with open(
            os.path.join(self.processed_dir, "pointcloud_metadata.pickle"), "wb"
        ) as f:
            pickle.dump(self.metadata, file=f)

    def select_y_from_metadata(self, selection: str, reload: bool = False) -> None:
        """
        It might happen that you want to process your dataset only once but then learn on different responses coming
        from the metadata object.

        Args:
            selection:
                The new selection of the :class:`pointcyto.data.metadata.MetaData` object.

            reload:
                If the selection is the same as previously and ``reload=False`` (default) then nothing happens.
                Only if ``reload=True`` the selection is "updated".
        Returns:
            None

        """
        if self.y_from_metadata_selection == selection and not reload:
            raise ValueError(
                "No new y selected as the selection is the same as the previous and reload=False"
            )
        self.metadata.selection = selection
        self.y_from_metadata_selection = selection
        # self.metadata.y automatically holds the response for the current selection
        self._data.y = torch.tensor(self.metadata.y)
        # Invalidate the cache. I am not sure what happens exactly within pytorch_geometric,
        # but this seems to work. See
        # https://pytorch-geometric.readthedocs.io/en/2.0.1/_modules/torch_geometric/data/in_memory_dataset.html?highlight=self._data_list%20%3D%20None%20#
        # https://pytorch-geometric.readthedocs.io/en/latest/_modules/torch_geometric/data/in_memory_dataset.html
        #   --> @data.setter does the same thing.
        self._data_list = None

    def __getitem__(self, item) -> Union["InMemoryPointCloud", Data]:
        data = super(InMemoryPointCloud, self).__getitem__(item)
        if not isinstance(item, int):
            data.metadata = data.metadata[item, :]

        return data



import inspect
from typing import List, Union

import flowkit as fk
import torch
from torch_geometric.data.dataset import Dataset


class FKtransform(object):
    """
    FKtransform is a class to access the transformations supplied by flowkit.

    I explain the parameters for all possible (2020-05-15) transformations from flowkit here also using information
    from Section 6.1 https://bioconductor.org/packages/devel/bioc/vignettes/flowCore/inst/doc/HowTo-flowCore.pdf
    Logicle and Hyperlog cannot be represented in closed form, so they are defined in terms of their inverse functions.
    https://www.bioconductor.org/packages/release/bioc/vignettes/flowCore/inst/doc/hyperlog.notice.html

    LinearTransform[t, a]

    .. math:`f(x) = \\frac{x + a}{b + a}`

    LogTransform[t, m]:

    .. math::

        f(x) = \\frac{1}{m} \\cdot \\log_{10}(x / t) +1

    HyperlogTransform[t, w, m, a]:
    (from https://github.com/whitews/FlowUtils/blob/master/flowutils/logicle_c_ext/logicle.c)

    .. math::

        f^{-1}(x) =


    LogicleTransform[t, w, m, a]:
    (from https://github.com/whitews/FlowUtils/blob/master/flowutils/logicle_c_ext/logicle.c)

    .. math::

        f^{-1}(x) =

    AsinhTransform[t, m, a]:

    .. math::

        f(x) =

        f^{-1}(x) =



    """

    def __init__(
        self,
        flowkit_transform_classname: str,
        feature_indices: Union[List[int], slice, int] = None,
        dataset_attribute: str = "pos",
        **kwargs
    ):
        # find all possible transforms from flowkit
        if flowkit_transform_classname not in fk.transforms.__all__:
            raise ValueError(
                flowkit_transform_classname + " does not exist in flowkit.transforms"
            )
        if flowkit_transform_classname == "RatioTransform":
            raise ValueError(
                "RatioTransform is special and not implemented because "
                + "it accesses pnn_labels from flowkit's sample class"
            )
        all_fk_transforms = {
            x[0]: x[1] for x in inspect.getmembers(fk.transforms, inspect.isclass)
        }
        # instantiate the chosen (flowkit_transform_classname) class with named arguments
        self.transform_class = all_fk_transforms[flowkit_transform_classname](
            transform_id=flowkit_transform_classname, **kwargs
        )

        self.attr = dataset_attribute
        self.feature_indices = feature_indices

    def __call__(self, data: Dataset):
        myattr = getattr(data, self.attr)
        if self.feature_indices is None:
            self.feature_indices = [*range(myattr.shape[1])]
        try:
            myattr[:, self.feature_indices] = self.transform_class.apply(
                myattr[:, self.feature_indices]
            )
        except AttributeError:
            #   File "/home/gugl/.conda_envs/ccc_optuna_III/lib/python3.8/site-packages/flowutils/transforms.py",
            #   line 231, in asinh
            #     data_copy = data.copy()
            # AttributeError: 'Tensor' object has no attribute 'copy'
            tmp = self.transform_class.apply(myattr[:, self.feature_indices].numpy())
            myattr[:, self.feature_indices] = torch.tensor(tmp)
        setattr(data, self.attr, myattr)
        return data



import inspect
import os


def find_dirname_above(
    dirpath: str,
    dirname: str = "tests",
    max_iterations: int = 30,
    verbose: bool = False,
    realpath: bool = True,
) -> str:
    """
    Goes directory-wise upwards starting at dirpath and
    0) current_directory = dirpath
    1) list content of current_directory and search for 'dirname'.
    2)  If found
            Return the path
        If not found
            current_directory = os.path.join(current_directory, '..')  # Ascend one directory, go to 1)

    Args:
        dirpath:
            current path
        dirname:
            what to search for
        max_iterations:
            Maximum number of going upwards in the directory-hierarchy (with os.pardir)
        verbose:
            To be or not to be verbose
        realpath:
            If true, return the os.path.realpath() of the returned path

    Returns:
        The first found path to the directory 'dirname'

    """

    TESTS_DIR = dirpath
    i = 0
    if verbose:
        print(TESTS_DIR)

    while dirname not in os.listdir(TESTS_DIR):
        TESTS_DIR = os.path.join(TESTS_DIR, os.pardir)
        if verbose:
            print(TESTS_DIR)
        i += 1
        if i > max_iterations:
            raise ValueError("There seems to be an error finding the test directory.")
    TESTS_DIR = os.path.join(TESTS_DIR, dirname)
    if realpath:
        TESTS_DIR = os.path.realpath(TESTS_DIR)
    if verbose:
        print(TESTS_DIR)

    return TESTS_DIR


def find_dirname_above_currentfile(
    dirname: str = "tests",
    max_iterations: int = 30,
    verbose: bool = False,
    realpath: bool = True,
) -> str:
    """
    Finds the given dirname above the file which called this function.

    Args:
        dirname:
            See find_dirname_above
        max_iterations:
            See find_dirname_above
        verbose:
            See find_dirname_above
        realpath:
            See find_dirname_above

    Returns:
        The first found path to the directory 'dirname'
    """
    # The following two lines extract the filename which called this function
    frame = inspect.stack()[1]
    filename = frame[0].f_code.co_filename

    return find_dirname_above(
        os.path.dirname(filename),
        dirname=dirname,
        max_iterations=max_iterations,
        verbose=verbose,
        realpath=realpath,
    )



import os
import re
import warnings
from typing import List, Union

import pandas as pd

from pointcyto.data.metadata_base import MetaDataBase


# This class is heavily based on torch_geometric.data.data
class MetaDataRegression(MetaDataBase):
    """
    A plain python object containing pheno information about a sample
    """

    def __init__(
        self,
        raw_filenames: List[str] = None,
        read_with: List[str] = "parse_by_ext",
        pheno: Union[dict, List[str], pd.DataFrame] = None,
        orig_dir: str = os.path.abspath("."),
        root: str = None,
        savetype: str = "torch",
        ignore_unequal_feature_names: bool = False,
        class_name: Union[dict, List[str], pd.DataFrame] = None,
    ):
        """

        Args:
            raw_filenames (List[str]):
                The raw filenames as list, relative path to orig_dir
            read_with:
                A list how the respective file should be read. Default: "ext" to read from file extension.
                2020-02-25: possible could be "parse_fcs", "parse_csv" or "sql"
            orig_dir:
                The directory where raw_filenames are inside
            root:
                The directory where the folders "raw/" and "processed/" should be inside.
                In "root/processed/", the processed filenames will be saved.

                If root is not given, root is defined as inside the orig_dir.
            savetype:
                How the processed files should be saved. Currently (2020-02-25) only implemented: "torch" and "csv".
                Future extension could be "sql" but then the conversion of raw_filenames -> processed_filenames
                must be handled extra. At the moment, each raw_filename is directly translated
                (by changing the extension) to a processed filename
            class_name:
                See MetaData, this is only to not break later MetaData handling and REPLACES `pheno`
                if given.
        """
        if pheno is None:
            pheno = class_name
        super(MetaDataRegression, self).__init__(
            raw_filenames=raw_filenames,
            read_with=read_with,
            pheno=pheno,
            orig_dir=orig_dir,
            root=root,
            savetype=savetype,
            ignore_unequal_feature_names=ignore_unequal_feature_names,
        )
        if self.pheno is not None:
            if self.pheno.shape[1] > 1:
                self.selection = self.pheno.columns.values[1]

    def add_pheno_single(
        self,
        new_pheno_column: Union[pd.DataFrame, List],
        col_name: str = None,
        replace_new_df_index: bool = True,
    ) -> None:
        """
        Adds a single new pheno column to the ``mymeta.pheno``. Make sure that if you supply a dataframe that it really
        is only one column, if you have more than one column use :meth:`add_pheno`.

        Args:
            new_pheno_column:
                A dataframe or list of elements.
            col_name:
                How should the new column be named? You always have to give this.
            replace_new_df_index:
                If you gave a dataframe in ``new_pheno_column``, should the existing index be removed?
                Examples where this is relevant are in
                :func:`tests.datasets.test_metadata.test_add_pheno_single_missing_index` and
                :func:`tests.datasets.test_metadata.test_add_pheno_single_switched_index`.

        Returns:

        """
        if not isinstance(new_pheno_column, pd.DataFrame):
            new_pheno_column = pd.DataFrame({col_name: new_pheno_column})

        if self._pheno is not None:
            if not new_pheno_column.index.equals(self._pheno.index):
                warnmessage_part1 = "new_pheno_column.index is NOT equal to the current self.pheno.index **"
                if replace_new_df_index:
                    new_pheno_column.index = self._pheno.index
                    # warnings.warn(
                    #     warnmessage_part1 +
                    #     '\n  ** but I was allowed to replace the index (what I did now with self.pheno.index)')
                else:
                    warnings.warn(
                        warnmessage_part1
                        + "\n  ** your given new_pheno_column is reordered and/or can contain Nan!"
                    )

        if self._pheno is None:
            self._pheno = new_pheno_column
        else:
            if len(new_pheno_column) != len(self):
                raise ValueError("New pheno must be equal length as the raw_filenames")
            if col_name is None:
                if isinstance(new_pheno_column, pd.DataFrame):
                    col_name = new_pheno_column.columns[0]
                if col_name is None:
                    col_name = "pheno_" + str(len(self._pheno.columns))
            elif col_name == 0:
                col_name = "pheno_" + str(len(self._pheno.columns))

            self._pheno[col_name] = new_pheno_column

    @property
    def y(self):
        """
        Return the currently selected class values as a list, depending on ``self.selection``

        Returns:
            List[Union[float, int]]
        """
        return self.pheno[self.selection].tolist()

    def append(self, new_listlike_dict: dict) -> None:
        """
        You can append new samples to the MetaData object using append.

        Args:
            new_listlike_dict:

        Returns:
            None

        Examples:
            See
            - :meth:`tests.datasets.test_metadata.test_metadata_append_only_filename`
            - :meth:`tests.datasets.test_metadata.test_metadata_append_with_sample_feature_names`
            - :meth:`tests.datasets.test_metadata.test_metadata_append_with_pheno`

        """
        super().append(new_listlike_dict=new_listlike_dict)
        try:
            tmp = pd.concat(
                [self.pheno, new_listlike_dict["pheno"]], sort=False, ignore_index=True
            )
            self._pheno = tmp
        except KeyError:
            # Add a row of NA as last row to _pheno
            self._pheno.loc[self._pheno.shape[0]] = pd.NA

    def __repr__(self):
        base_repr = super(MetaDataRegression, self).__repr__()
        if len(self) == 0:
            selected_y_range = "None"
        else:
            selected_y_range = f"[{min(self.y)}, {max(self.y)}]"
        base_repr = re.sub(
            pattern=r"(\nselection:.*\n)",
            repl=r"\1___AAAAAAAAAAAAAAAAAAAAAAAA___PLACEHOLDER___\n",
            string=base_repr,
        )
        return re.sub(
            "___AAAAAAAAAAAAAAAAAAAAAAAA___PLACEHOLDER___",
            "selected y range: {}".format(selected_y_range),
            base_repr,
        )



import os
import shutil
from unittest import TestCase

from pointcyto.io.meta_read_foldering import (
    find_classes,
    gen_csv_meta,
    gen_foldering_meta,
    gen_meta_filenames,
    make_dataset,
)
from pointcyto.io.utils import convert_class_id_names
from pointcyto.testutils.helpers import find_dirname_above

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
# Now build the base-test directory (necessary if you want to use tests/testdata)
TESTS_DIR = find_dirname_above(THIS_DIR)

pointcloud_toy_dataset = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_dataset"
)
pointcloud_toy_dataset_csvpheno = os.path.join(
    "testdata", "flowcytometry", "PointClouds_toy_phenocsv"
)


class Test(TestCase):
    def test_generate_foldering_meta(self):
        tmp = gen_foldering_meta(
            os.path.realpath(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw"))
        )
        repr(tmp)
        # Add additional optional parameters to MetaData
        tmp = gen_foldering_meta(
            os.path.realpath(os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")),
            root="asdf",
            savetype="csv",
        )
        repr(tmp)
        os.rmdir("asdf")

    def test_convert_class_id_names(self):
        orig_dir = os.path.join(TESTS_DIR, pointcloud_toy_dataset, "raw")
        classes_classtoidx = find_classes(orig_dir)
        fullpath_class_metadata = make_dataset(orig_dir, classes_classtoidx[1])
        relative_paths_class = [
            (path[len(orig_dir) + 1 :], class_id)
            for path, class_id in fullpath_class_metadata
        ]
        classes = [class_id for path, class_id in relative_paths_class]
        assert classes == [0, 0, 0, 1, 1, 1]
        assert convert_class_id_names(classes, classes_classtoidx[1], out="name") == [
            "A",
            "A",
            "A",
            "B",
            "B",
            "B",
        ]
        # The following cannot work because classes are already ids, so it regards "classes" as class_ids
        # With allow_missings it wont find anything.
        assert convert_class_id_names(
            classes, classes_classtoidx[1], out="id", allow_missings=True
        ) == [-1] * len(classes)
        # The following cannot work because classes are already ids, so it regards "classes" as class_ids
        try:
            convert_class_id_names(classes, classes_classtoidx[1], out="id")
        except ValueError:
            pass  # I inserted id and tried to convert to id, that must fail.

        converted_to_name = convert_class_id_names(
            classes, classes_classtoidx[1], out="name"
        )
        assert classes == convert_class_id_names(
            converted_to_name, classes_classtoidx[1], out="id"
        )

    def test_generate_meta_from_files_plus_pheno_csv(self):
        tmp = gen_csv_meta(
            os.path.realpath(os.path.join(TESTS_DIR, pointcloud_toy_dataset_csvpheno)),
            path_phenodata="samples_pheno.csv",
        )
        print(tmp.class_name)
        assert tmp.class_name == ["P0H.fcs", "P0Q.fcs", "PAG.fcs", "PAJ.fcs", "PAU.fcs"]
        tmp.selection = "label"
        assert tmp.class_name == [0, 0, 0, 0, 1]
        tmp.selection = "secondlabel"
        print(tmp.class_name)
        assert tmp.class_name == ["A", "A", "B", "B", "B"]
        repr(tmp)  # to test the printing without actually printing

    def test_subsetting(self):
        meta_AB = gen_csv_meta(
            os.path.join(TESTS_DIR, pointcloud_toy_dataset_csvpheno),
            path_phenodata="samples_pheno.csv",
        )
        meta_AB_train = meta_AB[0:2, :]
        meta_AB_train.rm_root(are_you_sure=True)
        print(meta_AB_train.processed_filenames)

    def test_gen_meta_filenames(self):
        meta_AB = gen_meta_filenames(
            os.path.realpath(
                os.path.join(
                    TESTS_DIR, "testdata", "flowcytometry", "info_in_filenames"
                )
            ),
            sep="_",
        )
        repr(meta_AB)
        # pd.Dataframe returns a "Series" when indexed
        assert meta_AB.pheno["column2"].tolist() == ["A", "B", "A"]

    def test_gen_meta_filenames_new_colnames(self):
        meta_AB = gen_meta_filenames(
            os.path.realpath(
                os.path.join(
                    TESTS_DIR, "testdata", "flowcytometry", "info_in_filenames"
                )
            ),
            sep="_",
            colnames=["sample", "smart_colname", "another_smart_colname"],
        )
        print(meta_AB)
        # # pd.Dataframe returns a "Series" when indexed
        # assert meta_AB.pheno["column2"].tolist() == ['A', 'B', 'A']

        with self.assertRaises(ValueError):
            meta_AB = gen_meta_filenames(
                os.path.realpath(
                    os.path.join(
                        TESTS_DIR, "testdata", "flowcytometry", "info_in_filenames"
                    )
                ),
                sep="_",
                colnames=[
                    "sample",
                    "a",
                    "b",
                    "more_colnames_than_necessary_NOT_allowed",
                ],
            )
            print(meta_AB)

        with self.assertRaises(ValueError):
            meta_AB = gen_meta_filenames(
                os.path.realpath(
                    os.path.join(
                        TESTS_DIR, "testdata", "flowcytometry", "info_in_filenames"
                    )
                ),
                sep="_",
                colnames=["sample", "less_colnames_than_necessary_NOT_allowed"],
            )
            print(meta_AB)

    def test_generate_meta_from_csv_wrong_path_delimiter(self):
        tmp = gen_csv_meta(
            os.path.realpath(os.path.join(TESTS_DIR, pointcloud_toy_dataset_csvpheno)),
            path_phenodata="samples_pheno_wrong_delimiter.csv",
        )
        tmp.selection = "label"
        assert tmp.class_name == [0, 0, 0, 0, 1]
        tmp.selection = "secondlabel"
        assert tmp.class_name == ["A", "A", "B", "B", "B"]
        repr(tmp)  # to test the printing without actually printing
        tmp = gen_csv_meta(
            os.path.realpath(os.path.join(TESTS_DIR, pointcloud_toy_dataset_csvpheno)),
            path_phenodata="samples_pheno_wrong_delimiter_2.csv",
        )
        tmp.selection = "label"
        assert tmp.class_name == [0, 0, 0, 0, 1]
        tmp.selection = "secondlabel"
        assert tmp.class_name == ["A", "A", "B", "B", "B"]
        repr(tmp)  # to test the printing without actually printing

    def tearDown(self) -> None:
        shutil.rmtree(
            os.path.join(
                TESTS_DIR,
                pointcloud_toy_dataset_csvpheno,
                os.pardir,
                "autoroot_PointClouds_toy_phenocsv",
            ),
            ignore_errors=True,
        )



import copy
import os
import re
import shutil
import warnings
from abc import abstractmethod
from typing import Dict, Iterable, List, Union

import pandas as pd
import torch
from pandas.core.series import Series


def size_repr(value):
    if torch.is_tensor(value):
        return list(value.size())
    elif isinstance(value, int) or isinstance(value, float):
        return [1]
    elif isinstance(value, list) or isinstance(value, tuple):
        return [len(value)]
    else:
        return value


def replace_file_extension(
    filename: str, to_ext: str = ".csv", regex: str = r"\.[^.]*$"
):
    return re.sub(regex, to_ext, filename)


# This class is heavily based on torch_geometric.data.data
class MetaDataBase(object):
    """
    A plain python object containing pheno information about a sample
    """

    def __init__(
        self,
        raw_filenames: List[str] = None,
        read_with: List[str] = "parse_by_ext",
        pheno: Union[dict, List[str], pd.DataFrame] = None,
        orig_dir: str = os.path.abspath("."),
        root: str = None,
        savetype: str = "torch",
        ignore_unequal_feature_names: bool = False,
    ):
        """

        Args:
            raw_filenames (List[str]):
                The raw filenames as list, relative path to orig_dir
            read_with:
                A list how the respective file should be read. Default: "ext" to read from file extension.
                2020-02-25: possible could be "parse_fcs", "parse_csv" or "sql"
            orig_dir:
                The directory where raw_filenames are inside
            root:
                The directory where the folders "raw/" and "processed/" should be inside.
                In "root/processed/", the processed filenames will be saved.

                If root is not given, root is defined as inside the orig_dir.
            savetype:
                How the processed files should be saved. Currently (2020-02-25) only implemented: "torch" and "csv".
                Future extension could be "sql" but then the conversion of raw_filenames -> processed_filenames
                must be handled extra. At the moment, each raw_filename is directly translated
                (by changing the extension) to a processed filename
            class_name:
                See MetaDataBase this is only to not break later MetaDataBasehandling and REPLACES `pheno`
                if given.
            ignore_unequal_feature_names:
                When building the Pointcloud, should unequal featurenames be ignored?
        """

        self._selection = None
        self._pheno = None
        self.savetype = savetype
        self.ignore_unequal_feature_names = ignore_unequal_feature_names

        # 1. Define all elements of this class
        # 1.1 List-like metadata, each element is one sample
        self.raw_filenames = raw_filenames

        if self.raw_filenames is not None:
            self.sample_feature_names = [
                None for _ in raw_filenames
            ]  # Is set outside of the Class
            # read_with should be a character vector of the function which
            #   1) reads the respective raw_filename
            #   2) Saves it in processed_filenames
            self.add_pheno_single(
                new_pheno_column=[read_with for _ in self.raw_filenames]
                if isinstance(read_with, str)
                else read_with,
                col_name="read_with",
            )
        else:
            self.sample_feature_names = []

        self.pheno = pheno
        if self.pheno is not None:
            if self.pheno.shape[1] > 1:
                self.selection = self.pheno.columns.values[1]
        # self._listlike_metadata is a set of listlike metadata elements
        self._listlike_metadata = {
            "raw_filenames",
            "processed_filenames",
            "sample_feature_names",
            "read_with",
            "y",
        }

        # 1.2 Single value or dictionary metadata, each value is relevant for all samples
        self.orig_dir = os.path.abspath(orig_dir)
        # If root is not given, root is defined as one folder above the orig_dir.
        only_orig_dir = os.path.basename(self.orig_dir)
        default_root_dir = os.path.realpath(
            os.path.join(self.orig_dir, os.pardir, "autoroot_" + only_orig_dir)
        )
        self.root = root if root is not None else default_root_dir
        os.makedirs(self.root, exist_ok=True)

    @property
    def selection(self) -> None:
        """
        When accessing :property:`~pointcyto.datametadata.MetaData.y` the returned values
        refer to a specific column of :property:`~pointcyto.datametadata.MetaData.pheno`.
        Which column should be used can be set by replacing the selection:

        .. code-block::

            self.selection = '<name_of_a_pheno_column>'
            self.selection = <number of a pheno column>

        Returns:
            None

        Examples:
            See :func:`tests.datasets.test_metadata.TestMetaData.metadata_testing`


        """
        return self._selection

    @selection.setter
    def selection(self, classname):
        if classname is None:
            self._selection = None
        elif isinstance(classname, int):
            self._selection = self.pheno.columns.values[classname]
        elif classname in self.pheno.columns.values:
            self._selection = classname
        else:
            raise KeyError(
                classname, "does not exist in self.pheno columns and is not integer."
            )

    @property
    def raw_filenames(self) -> List[str]:
        """
        Returns the raw filenames

        The ``raw_filenames.setter`` additionally generates the processed filenames using :meth:`proc_from_raw_filename`
        and stores it into ``self._processed_filenames`` so you can access it via (non-writable)
        :property:`processed_filenames`.

        Returns:
            A list of names for your files.
        """
        return self._raw_filenames

    @raw_filenames.setter
    def raw_filenames(self, filenames):
        self._raw_filenames = filenames
        if self.raw_filenames is not None:
            self._processed_filenames = [
                self.proc_from_raw_filename(file_x) for file_x in self.raw_filenames
            ]
        else:
            self._processed_filenames = None

    def proc_from_raw_filename(self, filename: str) -> str:
        """
        Each raw file might be processed and saved in a new file. This is mainly relevant for out-of-memory pointclouds
        as in-memory-pointclouds do not save each processed file one by one but one big file where all processed data
        is inside.

        The rawfile file-extension is replaced by a different extension based on ``self.savetype``.
        Args:
            filename:
                A string of the filename

        Returns:
            The filename string with the extension replaced.
        """
        if self.savetype == "torch":
            transformed = replace_file_extension(filename, ".pt")
        elif self.savetype == "csv":
            transformed = replace_file_extension(filename, ".csv")
        elif self.savetype == "sql":
            raise NotImplementedError
        else:
            raise NotImplementedError
        return transformed

    @property
    def processed_filenames(self) -> List[str]:
        return self._processed_filenames

    @property
    def pheno(self) -> pd.DataFrame:
        """
        The underlying phenodata for each raw_filename. Each raw_filename might have multiple values related to it,
        these values are organized in this dataframe.

        Note that the column ``read_with`` is always added here and set to the class constructor value ``read_with``

        Read-access to the phenodata by using self.pheno.
        Write-access by

        1. Setting the complete pheno `mymeta.pheno = <Dataframe or something which can be converted into one>`
        2. (Preferred) **adding** phenodata, see :meth:`add_pheno`

        Returns:
            The pheno dataframe
        """
        return self._pheno

    @pheno.setter
    def pheno(self, newpheno):
        if isinstance(newpheno, pd.DataFrame):
            newpheno_df = newpheno
        else:
            newpheno_df = pd.DataFrame(newpheno)
        self.add_pheno(newpheno_df)

    @property
    def n_pheno(self) -> int:
        """

        Returns:
            Returns how many phenodata-columns there are. Mind that ``read_with`` is generated by MetaData.__init__().
        """
        return len(self.pheno.columns)

    def add_pheno(
        self,
        new_pheno: Union[pd.DataFrame, Dict[str, List]],
        replace_new_df_index: bool = False,
    ) -> None:
        """
        Add new_pheno to the current phenodata.

        Args:
            new_pheno:
                Calls self.add_pheno_single() on every column of the given dataframe or creates a dataframe from a given
                ``{column_name: column_values_list}`` dictionary.
            replace_new_df_index:
                Forwarded to :meth:`add_pheno_single`
        Returns:
            None
        """
        if isinstance(new_pheno, Dict):
            new_pheno = pd.DataFrame(new_pheno)
        for colname in new_pheno.columns:
            self.add_pheno_single(
                new_pheno_column=new_pheno[colname],
                col_name=colname,
                replace_new_df_index=replace_new_df_index,
            )

    @abstractmethod
    def add_pheno_single(
        self,
        new_pheno_column: Union[pd.DataFrame, List],
        col_name: str = None,
        replace_new_df_index: bool = True,
    ) -> None:
        """
        Adds a single new pheno column to the ``mymeta.pheno``. Make sure that if you
        supply a dataframe that it really is only one column, if you have more than one
        column use :meth:`add_pheno`.

        Args:
            new_pheno_column:
                A dataframe or list of elements.
            col_name:
                How should the new column be named? You always have to give this.
            replace_new_df_index:
                If you gave a dataframe in ``new_pheno_column``, should the existing index be removed?
                Examples where this is relevant are in
                :func:`tests.datasets.test_metadata.test_add_pheno_single_missing_index` and
                :func:`tests.datasets.test_metadata.test_add_pheno_single_switched_index`.

        Returns:

        """

    @property
    def read_with(self) -> List[str]:
        """
        Extract the read_with column from the phenodata, used inside
        :class:`pointcyto.dataInMemoryPointCloud.InMemoryPointCloud`

        Returns:
            List[str] containing an identifier how the respective raw_filename should be read.
            Possibly in the future this could be a function, not only a string.
        """
        return self.pheno["read_with"].tolist()

    @property
    @abstractmethod
    def y(self):
        """
        Return the currently selected y values as a list, depending on ``self.selection``

        For classification: classIDs
        For regression: float values

        Returns:
            List[Union[float, int]]
        """

    @property
    def sample_feature_names(self) -> List[List[str]]:
        """
        Each sample might have feature names assigned to its raw_filename, you can set them here.

        Setter:
            Args:
                val:
                    Either supply the values(=val) as list(n_samples) of list(n_features) of str (single feature name)
                    or as a dictionary defining `{'feature_names': List[str], 'sample_number': List[int]}`
        """
        return self._sample_feature_names

    @sample_feature_names.setter
    def sample_feature_names(self, val: Union[dict, List[List[str]], List[str]]):
        if isinstance(val, dict):
            new_feature_names = val["feature_names"]
            sample_n = val["sample"]
            self._sample_feature_names[sample_n]: List[str] = new_feature_names
        else:
            if len(val) == len(self):
                self._sample_feature_names: List[List[str]] = val
            else:
                raise ValueError(
                    "List of feature names per sample has not the same length ("
                    + str(len(val))
                    + ")  as self ("
                    + str(len(self))
                    + ")"
                )

        self._set_feature_names()

    @property
    def feature_names(self) -> List[str]:
        """
        Holds the feature names over all samples in this MetaDataBaseobject. In contrast to
        :property:`sample_feature_names` this property does not hold the feature names *per sample*
        but *over all samples*.

        If the feature names are identical over all samples this is it. If the feature names are NOT identical over all
        samples a warning occurs AND this values becomes ``Feature names NOT identical over samples!``. You should
        probably take this warning very seriously because features of different samples might not be matching in order.
        If that is the case you have to fix your data first.

        Returns:
            A list of feature names over all samples.
        """
        return self._feature_names

    def _set_feature_names(self):
        self._feature_names_identical = True
        if len(self.sample_feature_names) == 0:
            allsamples_fn = None
        elif all(
            [
                self.sample_feature_names[0] == feature_names_sample_N
                for feature_names_sample_N in self.sample_feature_names
            ]
        ):
            allsamples_fn = self.sample_feature_names[0]
        else:
            self._feature_names_identical = False
            if not self.ignore_unequal_feature_names:
                warnings.warn("Feature names are not identical!", UserWarning)
            allsamples_fn = ["Feature names NOT identical over samples!"]
        self._feature_names = allsamples_fn

    @property
    def feature_names_identical(self):
        """
        Boolean value if the feature_names are identical over all samples.

        Returns:
            True/False
        """
        return self._feature_names_identical

    def _del_pheno_column(self, column: Union[str, int]) -> None:
        """
        Only removing a column of self.pheno is not enough to really purge the column as in ``classes_base_cmap``,
        ``class_colors`` and ``class_name_id_map`` still the values exist. This function completely removes those and
        the column in the phenodata.

        Args:
            column:
                Column name or index

        Returns:
            None
        """
        if isinstance(column, int):
            column = self.pheno.columns[column]
        if column == "read_with":
            raise IndexError("Cannot delete read_with column")
        self._pheno.drop(column, axis=1, inplace=True)
        if self.selection == column:
            self.selection = 0

    def __getitem__(self, key: Union[int, slice, tuple] = None):
        if isinstance(key, tuple) and len(key) == 2:
            new_metadata = copy.deepcopy(self)
            samples = key[0]
            pheno = key[1]
            if pheno is not None:
                if isinstance(pheno, int):
                    pheno_index = [pheno]
                elif isinstance(pheno, slice):
                    # See https://stackoverflow.com/questions/2936863/python-implementing-slicing-in-getitem
                    #   Eric Cousineau answer how slicing behaves
                    pheno_cn = list(new_metadata.pheno.columns.values)
                    start = pheno.start if pheno.start is not None else 0
                    stop = pheno.stop if pheno.stop is not None else len(pheno_cn)
                    step = pheno.step if pheno.step is not None else 1
                    pheno_index = [
                        pheno_cn.index(mypheno) if isinstance(mypheno, str) else mypheno
                        for mypheno in range(start, stop, step)
                    ]
                else:
                    pheno_cn = list(new_metadata.pheno.columns.values)
                    pheno_index = [
                        pheno_cn.index(mypheno) if isinstance(mypheno, str) else mypheno
                        for mypheno in pheno
                    ]
                if 0 not in pheno_index:
                    pheno_index = [0] + pheno_index
                del_pheno_indices = [
                    p_col
                    for p_col in range(len(self.pheno.columns))
                    if p_col not in pheno_index
                ]
                del_pheno_names = self.pheno.columns.values[del_pheno_indices]
                for single_pheno in del_pheno_names:
                    new_metadata._del_pheno_column(single_pheno)

            check_samples = samples != slice(None, None, None)
            # Checks *first* if check_samples is boolean
            #   If it is NOT, then it is probably something with greater length (e.g. a numpy array or a list)
            #   then go subset the filenames
            # If it is, then take its logical value.
            if not isinstance(check_samples, bool) or check_samples:
                try:
                    new_raw_filenames = []
                    for index, value in enumerate(samples):
                        if isinstance(value, bool):
                            if value:
                                new_raw_filenames.append(
                                    new_metadata.raw_filenames[index]
                                )
                        else:
                            new_raw_filenames.append(new_metadata.raw_filenames[value])
                except TypeError:  # then samples was no iterable
                    new_raw_filenames = new_metadata.raw_filenames[samples]

                if not isinstance(new_raw_filenames, list):
                    new_raw_filenames = [new_raw_filenames]

                new_metadata.raw_filenames = new_raw_filenames

                # Restrict phenodata but without updating the pheno keys (self._build_pheno_column())
                if isinstance(samples, int):
                    new_metadata._pheno = new_metadata._pheno.iloc[[samples]]
                    new_metadata.sample_feature_names = [
                        self.sample_feature_names[samples]
                    ]
                elif isinstance(samples, Series):
                    new_metadata._pheno = new_metadata._pheno.loc[samples, :]
                    new_metadata.sample_feature_names = [
                        self.sample_feature_names[index]
                        for index, value in enumerate(samples.values)
                        if value
                    ]
                else:
                    new_metadata._pheno = new_metadata._pheno.iloc[samples, :]
                    if isinstance(samples, Iterable):
                        if sum([isinstance(x, bool) for x in samples]) == 0:
                            new_metadata.sample_feature_names = [
                                self.sample_feature_names[index] for index in samples
                            ]
                        elif sum([isinstance(x, bool) for x in samples]) == len(
                            samples
                        ):
                            new_metadata.sample_feature_names = [
                                self.sample_feature_names[index]
                                for index, value in enumerate(samples)
                                if value
                            ]
                        else:
                            raise ValueError(
                                "Your given samples (metadata[samples, :]) are "
                                + "neither all boolean nor all not boolean. There seems to be a mistake. "
                            )
                    else:
                        new_metadata.sample_feature_names = self.sample_feature_names[
                            samples
                        ]

                # if isinstance(samples, int):
                #     new_metadata.sample_feature_names = [self.sample_feature_names[samples]]
                # else:
                #     print(samples)
                #     print([self.sample_feature_names[x] for x in samples])
                #     new_metadata.sample_feature_names = [self.sample_feature_names[x] for x in samples]

                # Now make a new auto_root_directory for fast later loading.
                # List current directory with root-directories inside.
                base_root_dir = os.path.basename(new_metadata.root)
                dir_with_root_dir = new_metadata.root[
                    0 : len(new_metadata.root) - len(base_root_dir) - 1
                ]
                base_root_dir_nonumber = re.sub(r"___[0-9]+$", "", base_root_dir)
                # https://stackoverflow.com/questions/7124778/how-to-match-anything-up-until-this-sequence-of-characters-in-a-regular-expres
                last_part_number = [
                    re.sub(r".+?(___([0-9]+))?$", r"\2", directory)
                    for directory in os.listdir(dir_with_root_dir)
                    if base_root_dir_nonumber in directory
                ]
                last_part_number = [
                    int(number) for number in last_part_number if number != ""
                ]
                last_part_number += [0]
                new_metadata.root = os.path.join(
                    dir_with_root_dir,
                    base_root_dir + "___" + str(max(last_part_number) + 1),
                )
                os.mkdir(new_metadata.root)

            return new_metadata
        else:
            samples = key
            return {
                key_x: getattr(self, key_x)[samples]
                for key_x in self._listlike_metadata
            }

    @abstractmethod
    def append(self, new_listlike_dict: dict) -> None:
        """
        You can append new samples to the MetaDataBaseobject using append.

        Args:
            new_listlike_dict:

        Returns:
            None

        Examples:
            See
            - :meth:`tests.datasets.test_metadata.test_metadata_append_only_filename`
            - :meth:`tests.datasets.test_metadata.test_metadata_append_with_sample_feature_names`
            - :meth:`tests.datasets.test_metadata.test_metadata_append_with_pheno`

        """
        if len(new_listlike_dict["raw_filenames"]) > 1:
            raise ValueError("Only supply a single raw_filename as list.")
        self._raw_filenames += new_listlike_dict["raw_filenames"]
        self._processed_filenames += [
            self.proc_from_raw_filename(new_listlike_dict["raw_filenames"][0])
        ]
        del new_listlike_dict["raw_filenames"]
        # Possible further keys: sample_feature_names
        try:
            self.sample_feature_names += [new_listlike_dict["sample_feature_names"]]
        except KeyError:
            pass

    @property
    def keys(self):
        r"""Returns all names of attributes."""
        keys = [key for key in self.__dict__.keys()]
        keys = [key for key in keys if key[:2] != "__" and key[-2:] != "__"]
        return keys

    def __len__(self):
        r"""Returns the number of raw files (read "samples")."""
        if self.raw_filenames is None:
            return 0
        else:
            return len(self.raw_filenames)

    def __contains__(self, key):
        r"""Returns :obj:`True`, if the attribute :obj:`key` is present in the
        data."""
        return key in self.keys

    def __iter__(self):
        r"""Iterates over all samples, so list elements in the "list-like metadata".
        Yielding a dictionary of all list-like metadata for the respective sample_n"""
        for sample_n in range(len(self.raw_filenames)):
            yield self[sample_n]

    def __call__(self, *keys):
        r"""Iterates over all samples given in *keys (int), so list elements in the "list-like metadata".
        Yielding a dictionary of all list-like metadata for the respective sample_n.

        If *keys is not given, iterates over all samples like __iter__()

        """
        if len(keys) > 0:
            for sample_n in keys:
                yield self[sample_n]
        else:
            for sample_n in range(len(self.raw_filenames)):
                yield self[sample_n]

    def __eq__(self, other):
        all_keys_present = all([key_x in other.keys for key_x in self.keys])
        if not all_keys_present:
            return False
        retval = True
        for key_x in self.keys:
            if isinstance(getattr(self, key_x, None), pd.DataFrame):
                retval = getattr(self, key_x, None).equals(getattr(other, key_x, None))
            elif getattr(self, key_x, None) != getattr(other, key_x, None):
                retval = False
        return retval

    def __repr__(self):
        feature_names = (
            ", ".join(self.feature_names) if self.feature_names is not None else "None"
        )
        if len(self) == 0:
            pheno_col_values = "None"
        else:
            pheno_col_values = self.pheno.columns.values

        return "\n".join(
            [
                "{} samples".format(len(self)),
                "pheno columns: {}".format(pheno_col_values),
                "selection: {}".format(str(self.selection)),
                "global feature names: {}".format(feature_names),
                "orig dir: {}".format(self.orig_dir),
                "root dir: {}".format(self.root),
                "savetype: {}".format(self.savetype),
            ]
        )

    def rm_root(self, are_you_sure: bool = False) -> None:
        """
        The PointCloud generation lateron will process the MetaDataBaseobject based on ``raw_filenames`` into
        ``processed_filenames`` into the ``self.root`` directory.

        If you want to delete this root directory you will loose all processed data (depending on your data
        this could take a bit of time) and if your ``root`` directory is the same as your ``orig_dir`` (do not do that)
        you might even loose your raw data.

        This is helpful because when subsetting a ``MetaDataBase` object, new root directories are generated
        and therefore it might be hard to track which ``MetaDataBase` object refers to which root-directory.

        Args:
            are_you_sure:
                Boolean, are you really sure you want to delete this?
                Introduced this parameter with default ``false`` that you cannot delete your data by error.

        Returns:
            None
        """
        if are_you_sure:
            if os.path.realpath(self.root) == os.path.realpath(self.orig_dir):
                raise ValueError(
                    "You specified root as orig_dir where the raw-data are located.\n"
                    + "I will not delete this for safety reasons."
                )
            shutil.rmtree(self.root)
        else:
            warnings.warn(
                "Did not remove root, specify are_you_sure as true. \n!!CAREFUL!!"
                + "\nYou will loose all of your processed data.\n"
                + "You might loose your original data if you specified "
            )
